type frag_t = (int * Num.t) list

type cstr_t =
| Direct of frag_t
| SplitEq of frag_t * frag_t

type t =
| Empty of frag_t
| Implies of (int * cstr_t) list
| Bind of (int * Cstr.t * cstr_t * t)

type cert_t = t

let cmp (i1,_) (i2,_) =
	match Pervasives.compare i1 i2 with
	| 0 -> invalid_arg "Cert.cmp"
	| n -> n

let sort a = List.sort cmp a

let isEqF (f1: frag_t) (f2: frag_t): bool =
	List.length f1 = List.length f2 &&
	List.for_all2 (fun (i1, n1) (i2, n2) -> i1 = i2 && Num.cmp n1 n2 = 0)
		(sort f1) (sort f2)

let mkF (init: (int * Num.t) list) =
	if List.for_all (fun (_, a) -> Num.cmpz a <> 0) init then
		sort init
	else
		invalid_arg "Cert.mkF"

let mapF (fn: Num.t -> Num.t) f =
	List.map (fun (i, n) -> (i, fn n)) f

let mergeF (l0: (frag_t * Num.t) list): frag_t =
	let l = List.map
		(fun (f, n) ->
			if Num.cmpz n = 0 then []
			else
				List.map (fun (i, a) -> (i, Num.mult n a)) f) l0
	in
	let rec merge2 f1 f2 =
		match f1, f2 with
		| [], t | t, [] -> t
		| (i1, n1)::t1, (i2, n2)::t2 ->
			if i1 < i2 then
				(i1, n1)::(merge2 t1 f2)
			else if i1 > i2 then
				(i2, n2)::(merge2 f1 t2)
			else
				let n = Num.add n1 n2 and t = merge2 t1 t2 in
				if Num.cmpz n = 0 then t else (i1, n)::t
	in
	List.fold_left merge2 [] (List.map sort l)

let inlineF (defs: (int * cstr_t) list) (f0: frag_t): frag_t =
	let _inline f1 (i0, n) =
	try
		let (_, d0) = List.find (fun (i1, _) -> i0 = i1) defs in
		match d0 with
		| Direct f2 -> mergeF [ f2, n; f1, Num.u ]
		| SplitEq (f2, f3) ->
			let d1 = if Num.cmpz n < 0 then (f2, n) else (f3, Num.neg n) in
			mergeF [ d1; f1, Num.u ]
	with
		Not_found -> (i0, n)::f1
	in
	List.fold_left _inline [] f0

let prF =
	function
	| [] -> "empty"
	| h::t ->
		let p (i, n) = (Num.pr n) ^ " * (" ^ (string_of_int i) ^ ")" in
		List.fold_left (fun s c -> s ^ " + " ^ (p c)) (p h) t

let inlineC (defs: (int * cstr_t) list) (typ: Cmp.t) (c0: cstr_t): cstr_t =
	match typ, c0 with
	| Cmp.Eq, Direct f -> SplitEq (inlineF defs f, inlineF defs (mapF Num.neg f))
	| Cmp.Le, Direct f | Cmp.Lt, Direct f ->Direct (inlineF defs f)
	| Cmp.Eq, SplitEq (f1, f2) -> SplitEq (inlineF defs f1, inlineF defs f2)
	| Cmp.Le, SplitEq _ | Cmp.Lt, SplitEq _ -> failwith "Cert.inlineC"

let prC =
	function
	| Direct f -> "Direct:\n\t" ^ (prF f)
	| SplitEq (f1, f2) -> "SplitEq:\n\t" ^ (prF f1) ^ "\n\t" ^ (prF f2)

(* XXX: nothing uses this. *)
let mk (i: (int * cstr_t) list) = Implies i

(* XXX: nothing uses this. *)
let add (c: cert_t) (i: int * cstr_t): cert_t =
	match c with
	| Empty _ | Bind _ -> invalid_arg "Cert.add"
	| Implies c -> Implies (i::c)

let rec pr =
	function
	| Empty f -> "Empty: " ^ (prF f)
	| Bind (id, _, ce1, ce2) -> Printf.sprintf "Bind %i: %s\nin\n%s" id (prC ce1) (pr ce2)
	| Implies ce0 ->
		match ce0 with
		| [] -> "Implies: empty"
		| h::t ->
			let p (i, c) = "(" ^ (string_of_int i) ^ "): " ^ (prC c) in
			List.fold_left (fun s c -> s ^ "\n" ^ (p c)) (p h) t
