type t = Cons.t list

type prop_t =
| Empty of Cert.frag_t
| Trivial
| Implied of int * Cert.cstr_t
| Check of Cons.t

type rel_t =
| NoIncl
| Incl of (int * Cert.cstr_t) list

type meet_t =
| MeetOk of (int * t)
| NewEqs of (Cons.t * t * (Cons.t * Cert.cstr_t) list)
| MeetBot of Cert.frag_t

type simpl_t =
| SimplMin of t
| SimplBot of Cert.frag_t

let nil = []

let synIncl es s c =
	(* XXX: This shouldn't be here. *)
	let extractCert id proof cons =
		let (d, w) = List.partition (fun (id, _) -> id = -1) cons in
		match d with
		| [_, a] ->
			let norm =
				let a1 = if Num.cmpz a < 0 then a else Num.neg a in
				(fun (id, a2) -> (id, Num.div a2 a1))
			in
			let w1 = List.append proof
				(List.map (fun (id, a1) -> (id, Num.neg a1)) w)
			in
			Implied (id, Cert.Direct (Cert.mkF (List.map norm w1)))

		| _ -> failwith "IneqSet.synIncl"
	in
	(* XXX: Is this check necessary? *)
	match Cstr.tellProp c.Cons.c with
	| Cstr.Trivial -> Trivial
	| Cstr.Contrad -> Empty (Cert.mkF [c.Cons.id, Num.u])
	| Cstr.Nothing ->

	(* XXX: Check that c.Cons.id is well-defined at every call site and get rid of the -1 dance. *)
	let c1 = EqSet.filter es (Cons.mk (-1) c.Cons.c) in
	match Cstr.tellProp c1.Cons.c with
	| Cstr.Trivial -> extractCert c.Cons.id [] c1.Cons.f
	| Cstr.Contrad -> Empty (List.map (fun (id, n) -> if id = -1 then (c.Cons.id, n) else (id, n)) c1.Cons.f)
	| Cstr.Nothing ->

	try
		let cI = List.find (fun c2 -> Cons.implies c2 c1) s in
		let ratio = (* XXX: should be done elsewhere *)
			match Vec.isomorph c1.Cons.c.Cstr.v cI.Cons.c.Cstr.v with
			| None -> failwith "IneqSet.synIncl"
			| Some r -> r
		in
		extractCert c.Cons.id [cI.Cons.id, ratio] c1.Cons.f
	with
		Not_found ->
			let (d, w) = List.partition (fun (id, _) -> id = -1) c1.Cons.f in
			match d with
			| [_, a] ->
				let f = Cert.mergeF [c.Cons.f, a; w,  Num.u] in
				Check { c1 with Cons.id = c.Cons.id; Cons.f = f }

			| _ -> failwith "IneqSet.synIncl"

let fromUnsat (w: Splx.witness_t) (id: int): Cert.frag_t =
	let (d, w1) = List.partition (fun (i, _) -> id = i) w in
	match d with
	| [] | _::_::_ -> invalid_arg "IneqSet.fromUnsat"
	| [_, a] ->
		let s = Num.cmpz a in
		if s = 0 then
			invalid_arg "IneqSet.fromUnsat"
		else
			(* XXX: This should be checked further,
			but I believe that the test below is never true.
			The intuition is that a constraint cannot be multiplied
			by a negative scalar. *)
			let norm =
				if s > 0 then
					fun (i, a1) -> (i, Num.div a1 (Num.neg a))
				else
					fun (i, a1) -> (i, Num.div a1 a)
			in
			Cert.mkF (List.map norm w1)

let isIncl b es s1 s2 =
	let rec _isIncl cert optSx =
	function
	| [] -> Incl cert
	| c::t ->

	match synIncl es s1 c with
	| Empty _ -> NoIncl
	| Trivial -> failwith "IneqSet.isIncl"
	| Implied (id, ce) -> _isIncl ((id, ce)::cert) optSx t
	| Check c1 ->

	let sx =
		match optSx with
		| Some sx -> sx
		| None -> Splx.mk b (List.map (fun c -> (c.Cons.id, c.Cons.c)) s1)
	in
	try
		let sx1 = Splx.add sx (-1, Cstr.compl c1.Cons.c) in
		let _ = Splx.check sx1 in
		NoIncl
	with
		Splx.Unsat w -> _isIncl ((c1.Cons.id, Cert.Direct (fromUnsat w (-1)))::cert) (Some sx) t
	in
	_isIncl [] None s2

let synAdd es s c =
	match synIncl es s c with
	| Trivial | Implied _ -> (c.Cons.id, s)
	| Empty _ -> failwith "IneqSet.synAdd"
	| Check c1 ->
		let s1 = c1::(List.filter (fun c2 -> not (Cons.implies c1 c2)) s) in
		(c1.Cons.id + 1, s1)

let trim s0 sx0 =
	let check c (s, sx) =
		let sx1 = Splx.compl sx c.Cons.id in
		try
			let _ = Splx.check sx1 in
			(c::s, sx)
		with
			Splx.Unsat _ -> (s, Splx.forget sx c.Cons.id)
	in
	(* XXX: Why is this fold_right? *)
	List.fold_right check s0 ([], sx0)

let findEqs s0 sx c0 =
	(* XXX: The first argument falls through. *)
	let checkEq s c =
		if c.Cons.c.Cstr.typ <> Cmp.Le then
			None
		else
			try
				let sx1 = Splx.stricten sx c.Cons.id in
				let _ = Splx.check sx1 in
				None
			with 
				Splx.Unsat w ->
					let eq = { c with Cons.c = { c.Cons.c with Cstr.typ = Cmp.Eq } } in
					let f1 = Cert.mkF [c.Cons.id, Num.u] and f2 = fromUnsat w c.Cons.id in
					Some (s, [eq, Cert.SplitEq (f1, f2)])
	in
	let rec checkEqs e s =
		function
		| [] -> (s, e)
		| h::t ->
			match checkEq t h with
			| None -> checkEqs e (h::s) t
			| Some (t1, e1) -> checkEqs (List.append e1 e) s t1
	in
	match checkEq s0 c0 with
	| None -> MeetOk (c0.Cons.id + 1, c0::s0)
	| Some (s1, e) ->
		let (s2, e1) = checkEqs e [] s1 in
		match e1 with
		| [] -> MeetOk (c0.Cons.id, s2)
		| _::_ -> NewEqs (c0, s2, e1)

let add b es s c =
	match c.Cons.c.Cstr.typ with
	| Cmp.Eq -> failwith "IneqSet.add"
	| Cmp.Le | Cmp.Lt ->

	match synIncl es s c with
	| Trivial | Implied _ -> MeetOk (c.Cons.id, s)
	| Empty f -> MeetBot f
	| Check c1 ->

	let s1 = List.filter (fun c2 -> not (Cons.implies c1 c2)) s in
	try
		let sx = Splx.mk b (List.map (fun c -> (c.Cons.id, c.Cons.c)) (c1::s1)) in
		let sx2 = Splx.check sx in
		let sx3 = Splx.compl sx2 c1.Cons.id in
		try
			let _ = Splx.check sx3 in
			let (s2, sx4) = trim s1 sx2 in
			findEqs s2 sx4 c1
		with
			Splx.Unsat _ -> MeetOk (c.Cons.id, s)
	with
		Splx.Unsat w -> MeetBot (Cert.inlineF [c1.Cons.id, Cert.Direct c1.Cons.f] w)

let isEq s1 s2 =
	let incl l1 = List.for_all (fun c2 ->
		List.exists (fun c1 -> Cstr.isInclSyn c1.Cons.c c2.Cons.c) l1)
	in
	incl s1 s2 && incl s2 s1

let rename s fromX toY = List.map (Cons.rename fromX toY) s

let pick msk s =
	let update v n p m =
		match Num.cmpz n with
		| 0 -> (v, p, m)
		| n when n < 0 -> (v, p + 1, m)
		| _ -> (v, p, m + 1)
	in
	let rec build m cnt ve =
		match m, cnt, ve with
		| Rtree.Nil, _, _ -> Rtree.Nil
		| _, rt, Rtree.Nil -> rt
		| Rtree.Sub (l1, n1, r1), Rtree.Nil, Rtree.Sub (l2, n2, r2) ->
			let l = build l1 Rtree.Nil l2 and r = build r1 Rtree.Nil r2 in
			let n =
				if n1 = None then
					(None, 0, 0)
				else
					update n1 n2 0 0
			in
			Rtree.Sub (l, n, r)

		| Rtree.Sub (l1,n1,r1), Rtree.Sub (l2,n2,r2), Rtree.Sub (l3,n3,r3) ->
			let l = build l1 l2 l3 and r = build r1 r2 r3 in
			let n =
				if n1 = None then
					n2
				else
					let (_, p, m) = n2 in
					update n1 n3 p m
			in
			Rtree.Sub (l, n, r)
	in
	let scores = List.fold_left (fun a c -> build msk a c.Cons.c.Cstr.v)
		Rtree.Nil s
	in
	let (optv, _) =
		let choose (best, sc) (v, p, m) =
			if v = None || p = 0 && m = 0 then
				(best, sc)
			else
				let nsc = p * m - (p + m) in
				if best = None then
					(v, nsc)
				else
					if nsc < sc then
						(v, nsc)
					else
						(best, sc)
		in
		Rtree.fold choose (None, -1) scores
	in
	optv

let trimSet b (id, s) =
	let sx = Splx.mk b (List.map (fun c -> (c.Cons.id, c.Cons.c)) s) in
	let (s1, _) = trim s (Splx.check sx) in
	(id, s1)

let simpl b es s =
	let rec filter s1 = function
	| [] ->
		(try
			let sx = Splx.mk b (List.map (fun c -> (c.Cons.id, c.Cons.c)) s1) in
			let (s2, _) = trim s1 (Splx.check sx) in
			SimplMin s2
		with
			Splx.Unsat w -> SimplBot w)
			
	| h::t ->
		match synIncl es s1 h with
		| Trivial | Implied _ -> filter s1 t
		| Empty f -> SimplBot f
		| Check h1 ->

		filter (h1::(List.filter (fun c -> not (Cons.implies h1 c)) s1)) t
	in
	filter [] (List.rev s)

let subst b es x id e s =
	let gen (id, s) c =
		let c1 =
			if Num.cmpz (Vec.get c.Cons.c.Cstr.v x) = 0 then
				{ c with Cons.id = id }
			else
				Cons.elimc id x e c
		in
		synAdd es s c1
	in
	trimSet b (List.fold_left gen (id, nil) s)

let fmElim b es x id s =
	let (pos, z, neg) = List.fold_left (fun (p, z, n) c ->
		match Num.cmpz (Vec.get c.Cons.c.Cstr.v x) with
		| 0 -> (p, c::z, n)
		| a when a < 0 -> (c::p, z, n)
		| _ -> (p, z, c::n)) ([], [], []) s
	in
	let (id1, zs) = List.fold_left
		(fun (id, s) c -> (id + 1, { c with Cons.id = id }::s)) (id, nil) z
	in
	let apply (id0, s0) c = List.fold_left
		(fun (id, s) c1 -> synAdd es s (Cons.elimc id x c c1)) (id0, s0) neg
	in
	trimSet b (List.fold_left apply (id1, zs) pos)

let fmElimM b es i msk s =
	let rec elim cnt i1 s1 =
		match pick msk s1 with
		| None ->
			if cnt = 0 then (* renumber even if nothing to project *)
				List.fold_right (fun c (id, s2) ->
					(id + 1, { c with Cons.id = id }::s2)) s1 (i, nil)
			else
				(i1, s1)

		| Some x ->
			let (i2, s2) = fmElim b es x i s1 in
			elim (cnt + 1) i2 s2
	in
	elim 0 i s

let setupJoin dup b reloc alpha idOff s =
	let apply (b1, reloc1, s1) c =
		let (b2, reloc2, c1) = Cons.setupJoin dup b1 reloc1 alpha idOff c in
		(b2, reloc2, c1::s1)
	in
	List.fold_left apply (b, reloc, nil) s

let extractc s = List.map Cons.extractc s
let initc s = List.map Cons.initc s

let isTop s = (s = [])
let list s = s

let pr b s = List.fold_left (fun str c -> str ^ (Cons.pr b c) ^ "\n") "" s

let prExt b s =
	List.fold_left (fun str c -> str ^ (Cons.prExt b c) ^ "\n") "" s
