module
Val =
	struct
	type t = { v: Num.t; d: Num.t }

	let of_num n = { v = n; d = Num.z }

	let pdelta n = { v = n; d = Num.u }

	let ndelta n = { v = n; d = Num.negU }

	let z = { v = Num.z; d = Num.z }

	let cmp v1 v2 =
		match Num.cmp v1.v v2.v with
		| 0 -> Num.cmp v1.d v2.d
		| n -> n

	let hasDelta v = (Num.cmpz v.d <> 0)

	let muln v n = {
		v = Num.mult v.v n;
		d = Num.mult v.d n
	}

	let divn v n = {
		v = Num.div v.v n;
		d = Num.div v.d n
	}

	let add v1 v2 = {
		v = Num.add v1.v v2.v;
		d = Num.add v1.d v2.d
	}

	let sub v1 v2 = {
		v = Num.sub v1.v v2.v;
		d = Num.sub v1.d v2.d
	}

	let adddelta v = { v with d = Num.add v.d Num.u }

	let subdelta v = { v with d = Num.sub v.d Num.u }

	let pr v = Printf.sprintf "%s + %s.delta" (Num.pr v.v) (Num.pr v.d)

	end

type bnd_t = { id: int; scale: Num.t; bv: Val.t }

type state_t = { v: Val.t; low: bnd_t option; up: bnd_t option }

type bck_t =
| Ok
| KoL of Val.t
| KoU of Val.t

type witness_t = (int * Num.t) list

type whatnext_t = NonBasic of Var.t | UnsatSplx of witness_t

type t = {
	mat: (Vec.t option) Rtree.t;
	state: state_t Rtree.t;
	b: Basis.t;
}

exception Sat of t
exception Unsat of witness_t

let prBnd = function
	| None -> "None"
	| Some b -> Printf.sprintf "Some {id = %i; scale = %s; bv = %s}" b.id (Num.pr b.scale) (Val.pr b.bv)

let prSt st = Printf.sprintf "{v = %s; low = %s; up = %s}" (Val.pr st.v) (prBnd st.low) (prBnd st.up)

let prState s =
	let rec _pr st ba =
		match st, ba with
		| Rtree.Nil, _ -> ""
		| _, Rtree.Nil -> failwith "Splx.prState"
		| Rtree.Sub (l1, xst, r1), Rtree.Sub (l2, x, r2) ->
			let lhs =
				match xst.low with
				| None -> "-inf <"
				| Some l -> (Val.pr l.bv) ^ " <="
			in
			let rhs =
				match xst.up with
				| None -> "< +inf"
				| Some u -> "<= " ^ (Val.pr u.bv)
			in
			let xstr = Printf.sprintf "%s (%s = %s) %s" lhs x (Val.pr xst.v) rhs in
			Printf.sprintf "%s\n%s%s" xstr (_pr l1 l2) (_pr r1 r2)
	in
	_pr s.state s.b.Basis.b

let prMat s =
	let rec _pr m ba =
		match m, ba with
		| Rtree.Nil, _ -> ""
		| _, Rtree.Nil -> failwith "Splx.prMat"
		| Rtree.Sub (l1, None, r1), Rtree.Sub (l2, _, r2) ->
			(_pr l1 l2) ^ (_pr r1 r2)
		| Rtree.Sub (l1, Some ve, r1), Rtree.Sub (l2, x, r2) ->
			let xstr = Printf.sprintf "%s: %s = 0\n" x (Vec.pr s.b ve) in
			Printf.sprintf "%s\n%s%s" xstr (_pr l1 l2) (_pr r1 r2)
	in
	_pr s.mat s.b.Basis.b

let pr s = Printf.sprintf "matrix:\n%s\nstate:\n%s\n" (prMat s) (prState s)

let mergeB st1 st2 =
	let l =
		match st1.low, st2.low with
		| l, None | None, l -> l
		| Some l1, Some l2 ->
			if Val.cmp l1.bv l2.bv < 0 then
				Some l2
			else
				Some l1
	and u =
		match st1.up, st2.up with
		| u, None | None, u -> u
		| Some u1, Some u2 ->
			if Val.cmp u1.bv u2.bv > 0 then
				Some u2
			else
				Some u1
	in
	match l, u with
	| Some l, Some u ->
		if Val.cmp l.bv u.bv > 0 then
			raise (Unsat [l.id, l.scale; u.id, u.scale])
		else
			{st1 with low = Some l; up = Some u}
	| None, None | None, Some _ | Some _, None -> {st1 with low = l; up = u}

let setVal ve st =
	let rec acc a = function
		| Rtree.Nil, _ | _, Rtree.Nil -> a
		| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, st2, r2) ->
			let a1 = Val.sub a (Val.muln st2.v n1) in
			acc (acc a1 (l1, l2)) (r1, r2)
	in
	acc Val.z (ve, st)

let fitBnd (st: state_t): Val.t * state_t =
	match st.low, st.up with
	| None, None -> (Val.z, st)
	| Some l, None ->
		if Val.cmp st.v l.bv < 0 then
			(Val.sub l.bv st.v, {st with v = l.bv})
		else
			(Val.z, st)
	| None, Some u ->
		if Val.cmp u.bv st.v < 0 then
			(Val.sub u.bv st.v, {st with v = u.bv})
		else
			(Val.z, st)
	| Some l, Some u ->
		if Val.cmp st.v l.bv < 0 then
			(Val.sub l.bv st.v, {st with v = l.bv})
		else if Val.cmp u.bv st.v < 0 then
			(Val.sub u.bv st.v, {st with v = u.bv})
		else
			(Val.z, st)

let rec incrupdate d xN m st =
	match m, st with
	| Rtree.Nil, rt -> rt
	| Rtree.Sub _, Rtree.Nil -> failwith "Splx.incrupdate"

	| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
		let l = incrupdate d xN l1 l2 and r = incrupdate d xN r1 r2 in
		let n =
			match n1 with
			| None -> n2
			| Some cons ->
				let n = Vec.get cons xN in
				{n2 with v = Val.sub n2.v (Val.muln d n)}
		in
		Rtree.Sub (l, n, r)

let mkBnd id (c: Cstr.t) (a: Num.t): state_t =
	let b: Num.t = Num.div c.Cstr.c a and aZ = Num.cmpz a in
	if aZ = 0 then
		invalid_arg "Splx.mkBnd"
	else
		match c.Cstr.typ with
		| Cmp.Eq ->
			let l = {id = id; scale = Num.inv (Num.neg a); bv = Val.of_num b}
			and u = {id = id; scale = Num.inv a; bv = Val.of_num b} in
			{v = Val.z; low = Some l; up = Some u}
		| Cmp.Le ->
			if aZ < 0 then
				let u = {id = id; scale = Num.inv a; bv = Val.of_num b} in
				{v = Val.z; low = None; up = Some u}
			else
				let l = {id = id; scale = Num.inv (Num.neg a); bv = Val.of_num b} in
				{v = Val.z; low = Some l; up = None}
		| Cmp.Lt ->
			if aZ < 0 then
				let u = {id = id; scale = Num.inv a; bv = Val.ndelta b} in
				{v = Val.z; low = None; up = Some u}
			else
				let l = {id = id; scale = Num.inv (Num.neg a); bv = Val.pdelta b} in
				{v = Val.z; low = Some l; up = None}

let add s (id, c) =
	match Rtree.fold (fun cnt n -> if Num.cmpz n = 0 then cnt else cnt + 1) 0 c.Cstr.v with
	| 0 -> invalid_arg "Splx.add"
	| 1 ->
		(* XXX: Can't we do without two tree traversals? *)
		(match Rtree.find (fun n -> if Num.cmpz n = 0 then None else Some n) c.Cstr.v with
		| None -> invalid_arg "Splx.add"
		| Some (x, a) ->
			let oBnd = Rtree.get {v = Val.z; low = None; up = None} s.state x in
			let delta, bnd = fitBnd (mergeB oBnd (mkBnd id c a)) in
			let nState = Rtree.set {v = Val.z; low = None; up = None} s.state x bnd in
			{s with state = incrupdate delta x s.mat nState})
	| _ ->
		let (nVar, b) = Basis.add s.b ("sx" ^ (string_of_int id)) in
		let nCons = Vec.set (Vec.neg c.Cstr.v) nVar Num.u in
		let nVState = {(mkBnd id c Num.u) with v = setVal nCons s.state} in
		let nMat = Rtree.set None s.mat nVar (Some nCons)
		and nState = Rtree.set {v = Val.z; low = None; up = None} s.state nVar nVState in
		{b = b; mat = nMat; state = nState}

let mk b (l: (int * Cstr.t) list) =
	let empty = {mat = Rtree.Nil; state = Rtree.Nil; b = b}
	in
	List.fold_left (fun s c -> add s c) empty l

let update m0 st0 =
	let updateval ve0 st0 n =
		let rec _upv a ve st =
			match ve, st with
			| Rtree.Nil, _ | _, Rtree.Nil -> a
			| Rtree.Sub (l1, _, r1), Rtree.Sub (l2, (None, _), r2) ->
				_upv (_upv a l1 l2) r1 r2

			| Rtree.Sub (l1, n, r1), Rtree.Sub (l2, (Some d, _), r2) ->
				_upv (_upv (Val.add a (Val.muln d n)) l1 l2) r1 r2
		in
		{ n with v = _upv n.v ve0 st0 }
	in
	let rec _up m st =
		match m, st with
		| _, Rtree.Nil -> Rtree.Nil
		| Rtree.Nil, Rtree.Sub (l, (_, n), r) ->
			Rtree.Sub (_up Rtree.Nil l, n, _up Rtree.Nil r)

		| Rtree.Sub (l1, None, r1), Rtree.Sub (l2, (_, n), r2) ->
			Rtree.Sub (_up l1 l2, n, _up r1 r2)

		| Rtree.Sub (l1, Some ve, r1), Rtree.Sub (l2, (_, n), r2) ->
			Rtree.Sub (_up l1 l2, updateval ve st0 n, _up r1 r2)
	in
	_up m0 st0

let pivot m xB xN =
	let xNCons =
		let xBCons =
			match Rtree.get None m xB with
			| None -> invalid_arg "Splx.pivot"
			| Some cons -> cons
		in
		Vec.div (Vec.get xBCons xN) xBCons
	in
	let elim = function
		| None -> None
		| Some cons -> Some (Vec.elim xN xNCons cons)
	in
	let m1 = Rtree.map elim (Rtree.set None m xB None) in
	Rtree.set None m1 xN (Some xNCons)

let _in_bound (b: state_t): bck_t =
	match b.low with
	| Some l when Val.cmp l.bv b.v > 0 -> KoL l.bv
	| Some _ | None ->
		match b.up with
		| Some u when Val.cmp b.v u.bv > 0 -> KoU u.bv
		| Some _ | None -> Ok

let picknbasic s fpos fneg xiv =
	let f w a st =
		match Num.cmpz a with
		| 0 -> UnsatSplx w
		| n when n < 0 -> fpos w a st
		| _ -> fneg w (Num.neg a) st
	in
	let rec find a0 ve st =
		match ve, st with
		| Rtree.Nil, _ -> UnsatSplx a0
		| Rtree.Sub _, Rtree.Nil -> NonBasic Var.XH
		| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
			match f a0 n1 n2 with
			| NonBasic _ as r -> r
			| UnsatSplx a1 ->
				match find a1 l1 l2 with
				| NonBasic v -> NonBasic (Var.XO v)
				| UnsatSplx a2 ->
					match find a2 r1 r2 with
					| NonBasic v -> NonBasic (Var.XI v)
					| UnsatSplx _ as r -> r
	in
	find [] xiv s.state

let pickbasic s =
	let tryincr w a st =
		match st.up with
		| None -> NonBasic Var.XH
		| Some u when Val.cmp u.bv st.v > 0 -> NonBasic Var.XH
		| Some u -> UnsatSplx ((u.id, Num.mult u.scale a)::w)
	in
	let trydecr w a st =
		match st.low with
		| None -> NonBasic Var.XH
		| Some l when Val.cmp l.bv st.v < 0 -> NonBasic Var.XH
		| Some l -> UnsatSplx ((l.id, Num.mult l.scale a)::w)
	in
	let f n st =
		match n with
		| None -> None
		| Some ve ->
			match _in_bound st with
			| Ok -> None
			| KoL l -> Some (picknbasic s trydecr tryincr ve, l)
			| KoU u -> Some (picknbasic s tryincr trydecr ve, u)
	in
	let rec find m st =
		match m, st with
		| Rtree.Nil, _ | _, Rtree.Nil -> None
		| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
			match f n1 n2 with
			| Some r -> Some (r, Var.XH)
			| None ->
				match find l1 l2 with
				| Some (r, va) -> Some (r, Var.XO va)
				| None ->
					match find r1 r2 with
					| Some (r, va) -> Some (r, Var.XI va)
					| None -> None
	in
	find s.mat s.state

let _check_body inst =
	match pickbasic inst with
	| None -> raise (Sat inst)
	| Some ((maybexj, xival), xi) ->
		match maybexj with
		| UnsatSplx w -> raise (Unsat w)
		| NonBasic xj ->
			let m = pivot inst.mat xi xj in
			let xiSt = Rtree.get {v = Val.z; low = None; up = None} inst.state xi in
			let st = Rtree.set {v=Val.z; low = None; up = None} inst.state xi {xiSt with v = xival} in
			{inst with mat = m; state = incrupdate (Val.sub xival xiSt.v) xi m st}

let presimpl s0 =
	let rec find ve st =
		let walk l1 l2 r1 r2 =
			match find l1 l2 with
			| Some (a, v) -> Some (a, Var.XO v)
			| None ->
				match find r1 r2 with
				| None -> None
				| Some (a, v) -> Some (a, Var.XI v)
		in
		match ve, st with
		| Rtree.Nil, _ -> None
		| Rtree.Sub (l, n, r), Rtree.Nil ->
			if Num.cmpz n = 0 then
				walk l Rtree.Nil r Rtree.Nil
			else
				Some (n, Var.XH)
		| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
			if Num.cmpz n1 <> 0 && n2.low = None && n2.up = None then
				Some (n1, Var.XH)
			else
				walk l1 l2 r1 r2
	in
	let (_, newm) =
		let rec simpl fl m =
			let recurse n fl l r =
				let (fll, nl) = simpl fl l in
				let (flr, nr) = simpl fll r in
				(flr, Rtree.Sub (nl, n, nr))
			in
			match m with
			| Rtree.Nil -> (fl, Rtree.Nil)
			| Rtree.Sub (l, None, r) -> recurse None fl l r
			| Rtree.Sub (l, Some ve, r) ->
				let nve = List.fold_right (fun f v -> f v) fl ve in
				match find nve s0.state with
				| None -> recurse (Some nve) fl l r
				| Some (a, va) ->
					recurse None ((Vec.elim va (Vec.div a nve))::fl) l r
		in
		simpl [] s0.mat
	in
	let nst = (* basic variables will be set by update *)
		Rtree.map (fun n ->
			match _in_bound n with
			| Ok -> (None, n)
			| KoL v | KoU v ->
				(Some (Val.sub n.v v), { n with v = v })) s0.state
	in
	{ s0 with mat = newm; state = update newm nst }

let check s0 =
	let rec check_loop s = check_loop (_check_body s) in
	try check_loop (presimpl s0)
	with Sat s -> s

let compl s id =
	let f s =
		match s.low, s.up with
		| Some l, _ when l.id = id ->
			let u1 = { l with bv = Val.subdelta l.bv } in
			Some (Val.sub u1.bv s.v, { v = u1.bv; low = None; up = Some u1 })

		| _, Some u when u.id = id ->
			let l1 = { u with bv = Val.adddelta u.bv } in
			Some (Val.sub l1.bv s.v, { v = l1.bv; low = Some l1; up = None })

		| _, _ -> None
	in
	match Rtree.find f s.state with
	| None -> invalid_arg "Splx.compl"
	| Some (v, (delta, b)) ->
		let st1 = Rtree.set {v = Val.z; low = None; up = None} s.state v b in
		{s with state = incrupdate delta v s.mat st1}

let stricten s id =
	let f s =
		match s.low, s.up with
		| Some l, None when l.id = id ->
			if Val.hasDelta l.bv then
				invalid_arg "Splx.stricten"
			else
				let l1 = { l with bv = Val.adddelta l.bv } in
				Some (Val.sub l1.bv s.v, { s with v = l1.bv; low = Some l1 })

		| Some l, Some u when l.id = id ->
			if Val.hasDelta l.bv then
				invalid_arg "Splx.stricten"
			else
				let l1 = { l with bv = Val.adddelta l.bv } in
				if Val.cmp l1.bv u.bv > 0 then
					raise (Unsat [l1.id, l1.scale; u.id, u.scale])
				else
					let s1 = { s with v = l1.bv; low = Some l1 } in
					Some (Val.sub l1.bv s.v, s1)

		| None, Some u when u.id = id ->
			if Val.hasDelta u.bv then
				invalid_arg "Splx.stricten"
			else
				let u1 = { u with bv = Val.subdelta u.bv } in
				Some (Val.sub u1.bv s.v, { s with v = u1.bv; up = Some u1 })

		| Some l, Some u when u.id = id ->
			if Val.hasDelta u.bv then
				invalid_arg "Splx.stricten"
			else
				let u1 = { u with bv = Val.subdelta u.bv } in
				if Val.cmp l.bv u1.bv > 0 then
					raise (Unsat [l.id, l.scale; u1.id, u.scale])
				else
					let s1 = { s with v = u1.bv; up = Some u1 } in
					Some (Val.sub u1.bv s.v, s1)

		| _, _ -> None
	in
	match Rtree.find f s.state with
	| None -> invalid_arg "Splx.stricten"
	| Some (v, (delta, b)) ->
		let st1 = Rtree.set {v = Val.z; low = None; up = None} s.state v b in
		{s with state = incrupdate delta v s.mat st1}

let forget s id =
	let f bnd =
		match bnd.low, bnd.up with
		| Some b, None | None, Some b when b.id = id ->
			Some (true, { bnd with low = None; up = None })

		| Some l, Some _ when l.id = id ->
			Some (false, { bnd with low = None })

		| Some _, Some u when u.id = id ->
			Some (false, { bnd with up = None })

		| _, _ -> None
	in
	match Rtree.find f s.state with
	| None -> invalid_arg "Splx.forget"
	| Some (v, (p, b)) ->
		let s1 = {s with state = Rtree.set {v = Val.z; low = None; up = None} s.state v b} in
		if p then
			presimpl s1 (* sledge hammer *)
		else
			s1

module Opt = struct
	type progressT = Unbnd | UpTo of Val.t | NoChange

	type dirT = Incr | Decr

	type actionT = Done | Move of (Var.t * dirT * progressT)

	type nextT = OptFinite of Val.t | OptUnbnd | GoOn of t

	type optT = Finite of Num.t | Sup of Num.t | Infty

	let prProgress = function
		| Unbnd -> "Unbnd"
		| UpTo v -> Printf.sprintf "UpTo %s" (Val.pr v)
		| NoChange -> "NoChange"
	
	let prDir = function
		| Incr -> "Incr"
		| Decr -> "Decr"
	
	let prAction b = function
		| Done -> "Done"
		| Move (v, d, bnd) -> Printf.sprintf "Move (%s, %s, %s)" (Basis.pr b v) (prDir d) (prProgress bnd)
	
	let prNext = function
		| GoOn s1 -> Printf.sprintf "GoOn\n%s\n" (pr s1)
		| OptUnbnd -> "OptUnbnd"
		| OptFinite v1 -> Printf.sprintf "OptFinite %s" (Val.pr v1)
	
	let prOpt = function
		| Finite n -> Printf.sprintf "Finite %s" (Num.pr n)
		| Sup n -> Printf.sprintf "Sup %s" (Num.pr n)
		| Infty -> "Infty"

	let tryDecr st =
		match st.low with
		| None -> Unbnd
		| Some l when Val.cmp l.bv st.v < 0 -> UpTo (Val.sub st.v l.bv)
		| Some _ -> NoChange

	let tryIncr st =
		match st.up with
		| None -> Unbnd
		| Some u when Val.cmp u.bv st.v > 0 -> UpTo (Val.sub u.bv st.v)
		| Some _ -> NoChange

	let pickNBasic z s =
		let rec find vec state =
			let walk l1 l2 r1 r2 =
				match find l1 l2 with
				| Move (x, dir, bnd) -> Move (Var.XO x, dir, bnd)
				| Done ->
					match find r1 r2 with
					| Move (x, dir, bnd) -> Move (Var.XI x, dir, bnd)
					| Done -> Done
			in
			match vec, state with
			| Rtree.Nil, _ -> Done
			| Rtree.Sub (l, n, r), Rtree.Nil ->
				let nSign = Num.cmpz n in
				if nSign = 0 then
					walk l Rtree.Nil r Rtree.Nil
				else
					Move (Var.XH, (if nSign < 0 then Decr else Incr), Unbnd)
			| Rtree.Sub (l1, n, r1), Rtree.Sub (l2, st, r2) ->
				let nSign = Num.cmpz n in
				if nSign = 0 then
					walk l1 l2 r1 r2
				else
					let (dir, tryFn) = if nSign < 0 then (Decr, tryDecr) else (Incr, tryIncr) in
					match tryFn st with
					| Unbnd | UpTo _ as b -> Move (Var.XH, dir, b)
					| NoChange -> walk l1 l2 r1 r2
		in
		match Rtree.get None s.mat z with
		| None -> invalid_arg "Splx.Opt.pickNBasic"
		| Some vec -> find (Vec.set vec z Num.z) s.state

	let pickBasic s xN xNBnd dir =
		let choose x1 bnd1 x2 bnd2 =
			match bnd1, bnd2 with
			| NoChange, NoChange | Unbnd, Unbnd -> (x1, bnd1)
			| NoChange, Unbnd | NoChange, UpTo _ | UpTo _, Unbnd -> (x1, bnd1)
			| Unbnd, UpTo _ | Unbnd, NoChange | UpTo _, NoChange -> (x2, bnd2)
			| UpTo d1, UpTo d2 ->
				if Val.cmp d1 d2 <= 0 then
					(x1, bnd1)
				else
					(x2, bnd2)
		in
		let (tryPos, tryNeg) =
			match dir with
			| Incr -> (tryDecr, tryIncr)
			| Decr -> (tryIncr, tryDecr)
		in
		let rec find m st =
			match m, st with
			| Rtree.Nil, _ -> None
			| _, Rtree.Nil -> Some (Var.XH, Unbnd)
			| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
				let nN =
					match n1 with
					| None -> Unbnd
					| Some cons ->
						let aN = Vec.get cons xN in
						match Num.cmpz aN with
						| 0 -> Unbnd
						| aSign ->
							match (if aSign < 0 then tryPos else tryNeg) n2 with
							| Unbnd | NoChange as bnd -> bnd
							| UpTo d -> UpTo (Val.divn d (Num.abs aN))
				in
				let maybeL = find l1 l2 and maybeR = find r1 r2 in
				match maybeL, maybeR with
				| None, None -> Some (Var.XH, nN)
				| Some (xL, nL), None -> Some (choose (Var.XO xL) nL Var.XH nN)
				| None, Some (xR, nR) -> Some (choose (Var.XI xR) nR Var.XH nN)
				| Some (xL, nL), Some (xR, nR) ->
					let (xS, nS) = choose (Var.XO xL) nL (Var.XI xR) nR in
					Some (choose xS nS Var.XH nN)
		in
		match find s.mat s.state with
		| None -> (xN, xNBnd)
		| Some (x, xBnd) -> choose xN xNBnd x xBnd

	let step z s =
		match pickNBasic z s with
		| Done ->
			let zSt = Rtree.get {v = Val.z; low = None; up = None} s.state z in
			OptFinite zSt.v
		| Move (xN, dir, b) ->
			let (xB, xNBnd) = pickBasic s xN b dir in
			match xNBnd with
			| Unbnd -> OptUnbnd
			| NoChange -> GoOn {s with mat = pivot s.mat xB xN}
			| UpTo delta ->
				let xNSt = Rtree.get {v = Val.z; low = None; up = None} s.state xN in
				let nSt =
					let upVal = match dir with Incr -> Val.add | Decr -> Val.sub in
					Rtree.set {v = Val.z; low = None; up = None} s.state xN {xNSt with v = upVal xNSt.v delta}
				in
				let m =
					if xB = xN then
						s.mat
					else
						pivot s.mat xB xN
				in
				let d = function Incr -> delta | Decr -> Val.sub Val.z delta in
				GoOn {s with mat = m; state = incrupdate (d dir) xN s.mat nSt}

	let setObj s obj =
		let (z, b1) = Basis.add s.b "sxZ" in
		let cons = Vec.set (Vec.neg obj) z Num.u in
		let zSt = {v = setVal cons s.state; low = None; up = None} in
		let nMat = Rtree.set None s.mat z (Some cons)
		and nState = Rtree.set {v = Val.z; low = None; up = None} s.state z zSt in
		(z, {mat = nMat; state = nState; b = b1})

	let max s0 obj =
		let (z, s1) = setObj s0 obj in
		let rec optimize s =
			match step z s with
			| GoOn s1 -> optimize s1
			| OptUnbnd -> Infty
			| OptFinite o ->
				match Num.cmpz o.Val.d with
				| 0 -> Finite o.Val.v
				| n when n > 0 -> Sup o.Val.v
				| _ -> failwith "Splx.Opt.max"
		in
		let rec check_loop s = check_loop (_check_body s) in
		try
			check_loop s1
		with
		| Sat s2 -> optimize s2

end
