(** This module implements a feasibility-only simplex algorithm.
The basis for the algorithm comes from the article "Integrating simplex with DPLL(T)"
published by Bruno Dutertre and Leonardo de Moura in 2006 as
an SRI International technical report.
The interface of this module needs a fair amount of user-friendliness. *)

module Val:
sig
	type t

	val of_num: Num.t -> t
	val pdelta: Num.t -> t
	val ndelta: Num.t -> t
	val z: t
	val cmp: t -> t -> int
	val hasDelta: t -> bool
	val muln: t -> Num.t -> t
	val divn: t -> Num.t -> t
	val add: t -> t -> t
	val sub: t -> t -> t
	val adddelta: t -> t
	val subdelta: t -> t
	val pr: t -> string
end

type bnd_t = { id: int; scale: Num.t; bv: Val.t }

type state_t = { v: Val.t; low: bnd_t option; up: bnd_t option }

type t = {
	mat: (Vec.t option) Rtree.t;
	state: state_t Rtree.t;
	b: Basis.t }

(*	The type of a contradiction witness.  This is the list of the non-zero
	coefficients to apply to the linear constraints of the instance, identified
	by an index number starting from zero and incremented for each newly added
	contraint.
	It describes a linear combination of the constraints of the problem yielding a
	trivially contradictory constraint.
*)
type witness_t = (int * Num.t) list

exception Unsat of witness_t

(** [add s (id, c)] adds contraints [c], identified by [id] to the simplex problem [s].
This function does not check that [id] uniquely identifies [c].
If [c] is trivial, [Invalid_argument "Splx.add"] is raised.
If [c] introduces a trivial contradiction in [s], [Unsat w] is raised,
where [w] exhibits contrary bounds of the system.
Trivial contradiction is detected when the provided bounds for a single variable are incompatible.
This means that [x < 1] and [x > 1] lead to a trivial contradiction detected by [add],
but not [x + y < 1] and [x + y > 1]. *)
val add: t -> int * Cstr.t -> t

(** [mk b l] builds a new simplex problem.
Basis [b] must be the basis the constraints in [l] refer to.
[mk] really just wraps calls to [add] and has the same limitations. *)
val mk: Basis.t -> (int * Cstr.t) list -> t

(** [check s] looks for an assignment of the variables which appear in the constraints of [s]
such that all the constraints of [s] are satisfied.
The function either terminates with a satisfying assignment or
raises [Unsat w] if it was able to prove that no such assignment exist.
The witness [w] exhibits contradictory bounds of the input system.
Due to a simplification step which occurs before the actual search,
the satisfying assignment cannot be extracted.
For the same reason, the addition of new constraints to the result of the [check] function,
following by a new call to [check] is not expected to work.
However, constraint manipulation using the functions [compl], [stricten] and [forget],
followed by a new call to [check] should work as expected. *)
val check: t -> t

(*	Complement the bound with given integer identifier.
	If multiple bounds have this identifier, one is arbitrarily chosen.
	Assumes the current point satisfies all constraints and bounds.
*)
val compl: t -> int -> t

(*	Make the bound with given integer identifier strict. *)
val stricten: t -> int -> t

(*	Remove the bound with the given integer identifier.
	If this is the only bound on the corresponding variable, project said
	variable from the constraints.
*)
val forget: t -> int -> t

(** Pretty-print an element of type [t].
The function [pr] wraps functions [prMat] and [prState]. *)
val pr: t -> string

(** [prMat s] pretty-prints the constraint matrix (field [mat]) of [s]. *)
val prMat: t -> string

(** [prState s] pretty-prints the bounds and values on variables (field [state]) of [s]. *)
val prState: t -> string

(** Pretty-print an element of type [bnd_t option]. *)
val prBnd: bnd_t option -> string

(** Pretty-print an element of type [state_t]. *)
val prSt: state_t -> string

(**/**)
(*	The following types and functions are private, hence the leading ``_''.
	They are used for diagnostic purposes.
*)

type bck_t =
| Ok
| KoL of Val.t
| KoU of Val.t

type whatnext_t = NonBasic of Var.t | UnsatSplx of witness_t

exception Sat of t

(** [mergeB st1 st2] computes the intersection of the bounds described by [st1] and [st2].
The [v] field of the result is equal to [st1.v].
If the interval resulting from the intersection is empty,
exception [Unsat w] is raised with [w] describing the offending bounds. *)
val mergeB: state_t -> state_t -> state_t

(** [setVal ve st] computes the amount by which the value of one of the variables involved in [ve]
(assuming it has a unit coefficient) should be increased so that the state [st] satisfies the constraint [ve]. *)
val setVal: Vec.t -> state_t Rtree.t -> Val.t

(** [fitBnd st = (d, st')] adjusts the value of the variable which state is described by [st], so that it fits its bounds.
It is assumed (but not enforced), that there is at least one possible value which fits the bounds.
The [fitBnd] function returns the new state [st'] with updated value and [d] = [st'.v - st.v]. *)
val fitBnd: state_t -> Val.t * state_t

(** [incrupdate v x m st] adjusts the value of the basic variables in [st] so that
the constraints [m] are still satisfied after a change [d] on the value of the non-basic variable [x] (new value is [x] + [d]). *)
val incrupdate: Val.t -> Var.t -> (Vec.t option) Rtree.t -> state_t Rtree.t -> state_t Rtree.t

(** [pivot m xB xN] transforms that contraint matrix [m] so that
the basic variable [xB] becomes non-basic and
the non-basic variable [xN] becomes basic.
The handling of variable values to preserve the invariant is left to the caller. *)
val pivot: Vec.t option Rtree.t -> Var.t -> Var.t -> Vec.t option Rtree.t

val _in_bound: state_t -> bck_t
val pickbasic: t -> ((whatnext_t * Val.t) * Var.t) option
val _check_body: t -> t
val presimpl: t -> t

(**/**)
module Opt: sig
(** This module builds on {! Splx} to offer an optimizing simplex. *)

(** A value of type [progressT] describes the change that can be operated on a given quantity.
The change does not specify whether this is an increase or a decrease.
In the case of [UpTo d], the change [d] is always positive. *)
	type progressT = Unbnd | UpTo of Val.t | NoChange

(** Describe the direction of a value change: increasing or decreasing. *)
	type dirT = Incr | Decr

(** During the optimization, a value of type [actionT] decribes the choice on the course of actions,
based on the contents of the objective line.
[Done] means that the optimal value for the objective function is reached.
[Move (v, d, p)] means that adjusting the value of the variable [v], in direction [d],
will increase the value of the objective function.
In the latter case, the bounds on [v] limit its change according to [p]. *)
	type actionT = Done | Move of (Var.t * dirT * progressT)

(** At the end of each pivoting step, decide whether the optimum is reached or
progress still needs to be made. *)
	type nextT = OptFinite of Val.t | OptUnbnd | GoOn of t

(** This is the result type of the optimization. *)
	type optT =
	| Finite of Num.t (** The optimized form can take values up to and including the given value. *)
	| Sup of Num.t (** The optimized form can take values up to and excluding the given value. *)
	| Infty (** The optimized form is unbounded. *)

(** [max s obj] maximizes the linear form [obj] under the constraints of [s].
The structure [s] may be built using the [add] function,
but [check] must not be called on it prior to calling [max].
If the constraints described by [s] are not satisfiable,
then [Unsat _] is raised in the same way as done by [check]. *)
	val max: t -> Vec.t -> optT

(** Pretty-print a value of type [progressT]. *)
	val prProgress: progressT -> string

(** Pretty-print a value of type [dirT]. *)
	val prDir: dirT -> string

(** Pretty-print a value of type [actionT].
The provided basis must be the one in which the non-basic variable is defined. *)
	val prAction: Basis.t -> actionT -> string

(** Pretty-print a value of type [nextT]. *)
	val prNext: nextT -> string

(** Pretty-print a value of type [optT]. *)
	val prOpt: optT -> string

(** The following function are internal functions which are exported for testing purposes only.
They maintain the invariant on the simplex structure,
which is that the value of all the variables should satisfy both their bounds and the constraints. *)

(** [pickNBasic z s] observes the objective line of the simplex tableau,
represented by the contraint of basic variable [z]
and decides what should be done. *)
	val pickNBasic: Var.t -> t -> actionT

(** [pickBasic s xN xNBnd dir] looks for the basic variable with which to pivot [xN].
The selected basic variable [xB] is the one which constrains the most the change of the value of [xN] in direction [dir].
The constraint on this change, [xNBnd1] is returned along with [xB].
The [xNBnd] argument is the constraint set on the change of the value of [xN] by the bounds of [xN].
If [xNBnd] is the most restrictive of all of the constraints imposed on the value of [xN], then [xB] = [xN]. *)
	val pickBasic: t -> Var.t -> progressT -> dirT -> Var.t * progressT

(** [step z s] does one optimization step:
it chooses a non-basic variable which can be used to increase the value of the goal,
it chooses a basic variable with which to pivot and finally pivots.
It notices that the optimum is reached only if it is reached in [s]. *)
	val step: Var.t -> t -> nextT

(** [setObj s obj] sets [obj] as the linear form to optimize in [s] and
returns the corresponding problem,
along with the basic variable associated with [obj]. *)
	val setObj: t -> Vec.t -> Var.t * t

end
