type t = Num.t Rtree.t

type vec_t = t

let nil = Rtree.Nil

let cut l n r: vec_t =
	if Num.cmpz n = 0 && l = Rtree.Nil && r = Rtree.Nil then
		Rtree.Nil
	else
		Rtree.Sub (l, n, r)

(* assume normalization *)
let rec isEq v1 v2: bool =
	match (v1, v2) with
	| (Rtree.Nil, Rtree.Nil) -> true
	| (Rtree.Sub _, Rtree.Nil) -> false
	| (Rtree.Nil, Rtree.Sub _) -> false
	| (Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2)) ->
		if Num.cmp n1 n2 = 0 then
			isEq l1 l2 && isEq r1 r2
		else
			false

let set ve0 va0 n0 =
	let z = (Num.cmpz n0 = 0) in
	let rec _set ve va =
		match ve, va with
		| Rtree.Nil, Var.XH when z -> Rtree.Nil
		| Rtree.Nil, Var.XH -> Rtree.Sub (Rtree.Nil, n0, Rtree.Nil)

		| Rtree.Sub (Rtree.Nil, _, Rtree.Nil), Var.XH when z -> Rtree.Nil
		| Rtree.Sub (l, _, r), Var.XH -> Rtree.Sub (l, n0, r)

		| Rtree.Nil, Var.XO t -> cut (_set Rtree.Nil t) Num.z Rtree.Nil
		| Rtree.Nil, Var.XI t -> cut Rtree.Nil Num.z (_set Rtree.Nil t)

		| Rtree.Sub (l, n, r), Var.XO t -> cut (_set l t) n r
		| Rtree.Sub (l, n, r), Var.XI t -> cut l n (_set r t)
	in
		_set ve0 va0

let mk (i: (Num.t * Var.t) list) =
	List.fold_left (fun v (n, var) -> set v var n) Rtree.Nil i

let get = Rtree.get Num.z

let rec add v1 v2: vec_t =
	match (v1, v2) with
	| (Rtree.Nil, Rtree.Nil) -> Rtree.Nil
	| (Rtree.Sub _, Rtree.Nil) -> v1
	| (Rtree.Nil, Rtree.Sub _) -> v2
	| (Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2)) ->
		cut (add l1 l2) (Num.add n1 n2) (add r1 r2)

let map f v0 =
	let rec _map v =
		match v with
		| Rtree.Nil -> Rtree.Nil
		| Rtree.Sub (l, n, r) -> cut (_map l) (f n) (_map r)
	in
	_map v0

let mult n v = map (Num.mult n) v

let div n0 v = map (fun n -> Num.div n n0) v

let elim (v: Var.t) (using: vec_t) (from: vec_t): vec_t =
	let n1 = get from v in
	match Num.cmpz n1 with
	| 0 -> from
	| _ ->
		let n2 = get using v in
		match Num.cmpz n2 with
		| 0 -> invalid_arg "Vec.elim"
		| _ -> add (mult (Num.neg n1) using) (mult n2 from)

let isomorph v1 v2 =
	let rec _iso optr v1 v2 =
		match v1, v2 with
		| Rtree.Nil, Rtree.Nil -> (optr, true)
		| Rtree.Nil, _ | _, Rtree.Nil -> (None, false)
		| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
			match Num.cmpz n1, Num.cmpz n2 with
			| 0, 0 ->
				let (optr1, res1) = _iso optr l1 l2 in
				if res1 then
					_iso optr1 r1 r2
				else
					(None, false)

			| 0,_ | _,0 -> (None, false)
			| _,_ ->
				let (res, r) =
					let ratio = Num.div n1 n2 in
					match optr with
					| None -> (true, Some ratio)
					| Some r -> (Num.cmp r ratio = 0, optr)
				in
				if res then
					let (_, res1) = _iso r l1 l2 in
					if res1 then
						_iso r r1 r2
					else
						(None, false)
				else
					(None, false)
	in
	let (ratio, _) = _iso None v1 v2 in ratio

let shift b0 ve reloc =
	let rec _shift b0 ve0 b ve reloc =
		let rshift l1 l2 l3 r1 r2 r3 x n re =
			let (b1, ve1, re1) =
				if Num.cmpz n = 0 then
					(b0, ve0, re)
				else
					match re with
					| Some va -> (b0, set ve0 va n, re)
					| None ->
						let (va, b1) = Basis.add b0 (x ^ "'1") in
						(b1, set ve0 va n, Some va)
			in
			let (b2, ve2, l3) = _shift b1 ve1 l1 l2 l3 in
			let (b3, ve3, r3) = _shift b2 ve2 r1 r2 r3 in
				(b3, ve3, Rtree.Sub (l3, re1, r3))
		in
		match b, ve, reloc with
		| Rtree.Nil, _, _ | _, Rtree.Nil, _ -> (b0, ve0, reloc)
		| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2), Rtree.Nil ->
			rshift l1 l2 Rtree.Nil r1 r2 Rtree.Nil n1 n2 None

		| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2),
			Rtree.Sub (l3, n3, r3) -> rshift l1 l2 l3 r1 r2 r3 n1 n2 n3
	in
		_shift b0 Rtree.Nil b0.Basis.b ve reloc

let rec mult_t v1 v2: vec_t =
	match (v1, v2) with
	| (Rtree.Nil, Rtree.Nil) -> Rtree.Nil
	| (Rtree.Sub _, Rtree.Nil) -> Rtree.Nil
	| (Rtree.Nil, Rtree.Sub _) -> Rtree.Nil
	| (Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2)) ->
		cut (mult_t l1 l2) (Num.mult n1 n2) (mult_t r1 r2)

let neg = map Num.neg

let gcd v =
	let gcd = Rtree.fold (fun g a ->
		if Num.cmpz a = 0 then
			g
		else
			match g with
			| None -> Some (Num.toZ (Num.abs a))
			| Some g -> Some (Num.gcd g a)) None v
	in
	match gcd with
	| None -> Num.u
	| Some (nGcd, dGcd) -> Num.ofZ dGcd nGcd

let pr b v =
	let s = Rtree.pr " + "
		(fun x a -> if Num.cmpz a = 0 then "" else (Num.pr a) ^ "." ^ x)
			b.Basis.b v
	in
	if s = "" then "0" else s
