(* EqSet *)
let (x, b) = Basis.add Basis.nil "x"
let (y, b) = Basis.add b "y"
let (z, b) = Basis.add b "z"
let (t, b) = Basis.add b "t"

let mkc t v c =
	Cstr.mk t (List.map (fun (n, x) -> (Num.mk n, x)) v) (Num.mk c)

let eq = mkc Cmp.Eq and le = mkc Cmp.Le and lt = mkc Cmp.Lt

let mask l =
	List.fold_left (fun m x -> Rtree.set None m x (Some x)) Rtree.Nil l

let mk l = (* does not check redundancy *)
	let (_, s) = List.fold_left
		(fun (i, s) (x, c) -> (i + 1, EqSet.addf s (x, Cons.mk i c)))
			(1, EqSet.nil) l
	in s

let optxpr =
	function
	| None -> "None"
	| Some x -> "Some " ^ (Basis.pr b x)

let prProp = function
| Cstr.Trivial -> "trivial"
| Cstr.Contrad -> "contrad"
| Cstr.Nothing -> "nothing"

(* EqSet.filter *)
let filter_ts =
	let chk_no (t, _, c, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let c1 = EqSet.filter s c in
		let vars = List.map (fun (x, _) -> x) s in
		let a =
			let zcoef x = Num.cmpz (Vec.get c1.Cons.c.Cstr.v x) = 0 in
			List.for_all zcoef vars
		in
		let e =
			"some variable defined in s has non-zero coefficient in c1\n" ^
			(Cons.prExt b c1) ^ "\n"
		in
		OUnit.assert_bool e a))
	in
	let chk_id (t, _, c, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let c1 = EqSet.filter s c in
		let e =
			"c.id: " ^ (string_of_int c.Cons.id) ^
			", c1.id " ^ (string_of_int c1.Cons.id)
		in
		OUnit.assert_bool e (c.Cons.id = c1.Cons.id)))
	in
	let chk_prop (t, p, c, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let c1 = EqSet.filter s c in
		let p1 = Cstr.tellProp c1.Cons.c in
		let e = (prProp p) ^ " != " ^ (prProp p1) in
		OUnit.assert_bool e (p = p1)))
	in
	let chk_lin (t, _, c, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let c1 = EqSet.filter s (Cons.initc c) in
		let (_, cert) = Cons.extractc c1 in
		let pairs =
			List.map (fun c -> (c.Cons.id, c.Cons.c)) (c::(EqSet.list s))
		in
		let a =
			if Cstr.tellProp c1.Cons.c = Cstr.Contrad then
				true
			else
				Glue.check pairs [1, c1.Cons.c] (Cert.Implies [1, cert])
		and e = "cannot check with:\n" ^ (Cert.prC cert) in
		OUnit.assert_bool e a))
	in
	let mk l =
		let add (i, s) c =
			match EqSet.add s (Cons.mk i c) with
			| EqSet.Bottom _ -> failwith "EqSet_t.tryDefs_ts"
			| EqSet.Ok (i1, s1) -> (i1, s1)
		in
		let (_, s) = List.fold_left add (0, EqSet.nil) l in
		s
	in
	let tcs = [
		"nil0", Cstr.Nothing, Cons.mk 0 (eq [1, x] 0), mk [];
		"nil1", Cstr.Nothing, Cons.mk 1 (eq [1, x] 1), mk [];
		"nil2", Cstr.Nothing, Cons.mk 2 (eq [1, x; 2, y] 0), mk [];
		"one0", Cstr.Nothing, Cons.mk 3 (eq [2, x; 3, y] 1), mk [
			eq [1, y] 1 ];

		"incl0", Cstr.Trivial, Cons.mk 4 (eq [2, x; 3, y] 1), mk [
			eq [1, x; 1, y] 1;
			eq [1, y] (-1) ];

		"incl1", Cstr.Trivial, Cons.mk 4 (le [2, x; 3, y] 1), mk [
			eq [1, x; 1, y] 1;
			eq [1, y] (-1) ];

		"triv0", Cstr.Trivial, Cons.mk 4 (le [] 1), mk [
			eq [1, x] 1;
			eq [1, y] 2 ];

		"contrad0", Cstr.Contrad, Cons.mk 5 (eq [2, x; 1, z] 1), mk [
			eq [1, x; 1, z] 0;
			eq [1, x] 2 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("no", OUnit.TestList (List.map chk_no tcs));
		OUnit.TestLabel ("id", OUnit.TestList (List.map chk_id tcs));
		OUnit.TestLabel ("prop", OUnit.TestList (List.map chk_prop tcs));
		OUnit.TestLabel ("lin", OUnit.TestList (List.map chk_lin tcs))
	]

(* EqSet.add *)
let add_ts =
	let mk l =
		let add (i, s) c =
			match EqSet.add s (Cons.mk i c) with
			| EqSet.Bottom _ -> failwith "EqSet_t.add_ts"
			| EqSet.Ok (i1, s1) -> (i1, s1)
		in
		List.fold_left add (0, EqSet.nil) l
	in
	let mkF l = Cert.mkF (List.map (fun (i, n) -> (i, Num.mk n)) l) in
	let chkBot (t, bot, _, e, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let (res, msg) =
			let (i, s1) = mk s in
			let e1 = Cons.mk i e in
			match EqSet.add s1 e1 with
			| EqSet.Ok _ -> (not bot, "should be bottom")
			| EqSet.Bottom f ->
				if not bot then
					(false, "should be non-empty")
				else
					let list = List.map (fun c -> (c.Cons.id, c.Cons.c)) (e1::(EqSet.list s1)) in
					let msg =
						Printf.sprintf "cannot check certificate\n%s%s\n%s\n"
						(EqSet.prExt b s1) (Cons.prExt b e1) (Cert.prF f)
					in
					(Glue.check list [] (Cert.Empty f), msg)
		in
		OUnit.assert_bool msg res))
	in
	let chkFrag (t, bot, f, e, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		if bot then () else
		let (i, s1) = mk s in
		let e1 = Cons.mk i e in
		match EqSet.add s1 e1 with
		| EqSet.Bottom _ -> OUnit.assert_bool "should be non-empty" false
		| EqSet.Ok (i1, s2) ->
			if i = i1 then () else (* XXX: we should check that e1 ~ e2 *)
			let e2 = List.find (fun e -> e.Cons.id = i) (EqSet.list s2) in
			OUnit.assert_equal ~cmp:Cert.isEqF ~printer:Cert.prF (mkF f) e2.Cons.f))
	in
	let tcs = [ (* XXX: Should frag and bot test cases be split? *)
		"contrad0", true, [], eq [] 1, [];
		"contrad1", true, [], eq [1, x] 1, [ eq [1, x] 0 ];
		"ok0", false, [1, 1], eq [1, x] 1, [ eq [1, y] 1 ];
		"ok1", false, [1, -1], eq [-1, x] 0, [ eq [1, y] 1 ];
		"ok2", false, [0, -1; 1, 1], eq [1, x; 1, y] 0, [ eq [1, y] 1 ];
		"ok3", false, [0, -1; 1, -1], eq [-1, x; -1, y] 0, [ eq [1, y] 1 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("bot", OUnit.TestList (List.map chkBot tcs));
		OUnit.TestLabel ("frag", OUnit.TestList (List.map chkFrag tcs))
	]

(* EqSet.pick *)
let pick_ts =
	let chk (t, msk, c, r) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let r1 = EqSet.pick msk c in
		let e = "expected " ^ (optxpr r) ^ " but got " ^ (optxpr r1) in
		OUnit.assert_bool e (r = r1)))
	in
	let tcs = [
		"nil0", mask [], Cons.mk 0 (eq [] 0), None;
		"nil1", mask [x], Cons.mk 0 (eq [] 0), None;
		"nil2", mask [y], Cons.mk 0 (eq [] 0), None;
		"nil3", mask [], Cons.mk 0 (eq [1, x] 0), None;
		"diff0", mask [x], Cons.mk 0 (eq [1, y] 0), None;
		"one0", mask [x], Cons.mk 0 (eq [1, x] 0), Some x;
		"one1", mask [x], Cons.mk 0 (eq [1, x; 1, y] 0), Some x;
		"one2", mask [y], Cons.mk 0 (eq [1, x; 1, y] 0), Some y;
		"m0", mask [x; y], Cons.mk 0 (eq [1, y] 0), Some y;
		"m1", mask [x; y], Cons.mk 0 (eq [1, y; 1, z] 0), Some y;

		(* implementation detail *)
		"m2", mask [x; y], Cons.mk 0 (eq [1, x; 1, y] 0), Some x
	] in
	OUnit.TestList (List.map chk tcs)

(* EqSet.subst *)
let subst_ts =
	let chk_id (t, x, i, i1, c, s, _) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let c1 = Cons.mk 0 c in
		let (i2, _) = EqSet.subst x i c1 s in
		let e =
			"expected " ^ (string_of_int i1) ^
			" but got " ^ (string_of_int i2)
		in
		OUnit.assert_bool e (i1 = i2)))
	in
	let chk_cons (t, x, i, _, c, s, s1) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let c1 = Cons.mk 0 c in
		let (_, s2) = EqSet.subst x i c1 s in
		let e =
			"expected:\n" ^  (EqSet.prExt b s1) ^ "but got:\n" ^ (EqSet.prExt b s2)
		in
		OUnit.assert_bool e (EqSet.isEq s1 s2)))
	in
	let tcs = [
		"nil0", x, 0, 0, eq [1, x] 0, mk [], mk [];
		"nil1", y, 0, 0, eq [1, x] 0, mk [], mk [];
		"nil2", y, 1, 1, eq [1, x] 0, mk [], mk [];
		"one0", x, 0, 1, eq [1, x] 0,
			mk [y, eq [1, x; 1, y] 1],
			mk [y, eq [1, y] 1];

		"one1", x, 1, 2, eq [1, x] 0,
			mk [y, eq [1, x; 1, y] 1],
			mk [y, eq [1, y] 1];

		"no0", x, 0, 1, eq [1, x] 0, mk [y, eq [1, y] 1], mk [y, eq [1, y] 1];
		"m0", x, 0, 2, eq [1, x] 0, mk [
			z, eq [1, z] 1;
			y, eq [1, x; 1, y] 1
		], mk [
			z, eq [1, z] 1;
			y, eq [1, y] 1 ];
	] in
	OUnit.TestList [
		OUnit.TestLabel ("id", OUnit.TestList (List.map chk_id tcs));
		OUnit.TestLabel ("id", OUnit.TestList (List.map chk_cons tcs))
	]

(* EqSet.tryDefs *)
let tryDefs_ts =
	let chk_int (t, _, i0, msk, s) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let (_, i1, s1) = EqSet.tryDefs i0 msk s in
		let a = List.for_all
			(fun c -> i0 <= c.Cons.id && c.Cons.id < i1) (EqSet.list s1)
		in
		let e =
			"not all ids between " ^ (string_of_int i0) ^
			" and " ^ (string_of_int i1) ^ "\n" ^ (EqSet.prExt b s1)
		in
		OUnit.assert_bool e a))
	in
	let chk_dup (t, _, i0, msk, s) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let (_, _, s1) = EqSet.tryDefs i0 msk s in
		let ids = List.sort Pervasives.compare
			(List.map (fun c -> c.Cons.id) (EqSet.list s1))
		in
		let rec chk i1 =
			function
			| [] -> true
			| i2::l ->
				if i1 = i2 then
					false
				else
					chk i2 l
		in
		let a =
			match ids with
			| [] -> true
			| i::l -> chk i l
		in
		let e = "found duplicate ids\n" ^ (EqSet.prExt b s1) in
		OUnit.assert_bool e a))
	in
	let chk_res (t, r, i0, msk, s) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let (def, _, _) = EqSet.tryDefs i0 msk s in
		let a = if def = None then false else true in
		OUnit.assert_bool "bad conclusion" (a = r)))
	in
	let chk_def (t, _, i0, msk, s) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let (def, _, _) = EqSet.tryDefs i0 msk s in
		match def with
		| None -> ()
		| Some (_, x) ->
		OUnit.assert_bool "not found" (List.exists (fun (x1, _) -> x = x1) s)))
	in
	let chk_nox (t, _, i0, msk, s) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let (def, _, s1) = EqSet.tryDefs i0 msk s in
		match def with
		| None -> ()
		| Some (_, x) ->
		OUnit.assert_bool "x remaining" (List.for_all (fun (_, c) ->
			Num.cmpz (Vec.get c.Cons.c.Cstr.v x) = 0) s1)))
	in
	let mk l =
		let add (i, s) c =
			match EqSet.add s (Cons.mk i c) with
			| EqSet.Bottom _ -> failwith "EqSet_t.tryDefs_ts"
			| EqSet.Ok (i1, s1) -> (i1, s1)
		in
		let (_, s) = List.fold_left add (0, EqSet.nil) l in
		s
	in
	let tcs = [
		"nil0", false, 0, mask [], mk [];
		"nil1", false, 0, mask [x], mk [];
		"one0", true, 0, mask [x], mk [eq [1, x] 0];
		"one0", true, 0, mask [x], mk [
			eq [1, x] 0;
			eq [1, x; 1, y] 0 ];

		"no0", false, 2, mask [z], mk [
			eq [1, x; 1, z] 0;
			eq [1, y] 0 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("int", OUnit.TestList (List.map chk_int tcs));
		OUnit.TestLabel ("dup", OUnit.TestList (List.map chk_dup tcs));
		OUnit.TestLabel ("res", OUnit.TestList (List.map chk_res tcs));
		OUnit.TestLabel ("def", OUnit.TestList (List.map chk_def tcs));
		OUnit.TestLabel ("nox", OUnit.TestList (List.map chk_nox tcs))
	]

(* EqSet.trySubstM *)
let trySubstM_ts =
	let chk_var (t, i, x, msk, s) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let (optx1, _, _) = EqSet.trySubstM i msk s in
		match optx1 with
		| None ->
			let e = "expected " ^ (optxpr x) ^ " but got None" in
			OUnit.assert_bool e (x = None)

		| Some (_, x1) ->
			let e =
				"expected " ^ (optxpr x) ^ " but got " ^ (optxpr (Some x1))
			in
			OUnit.assert_bool e (x = Some x1)))
	in
	let chk_nox (t, i, x, msk, s) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let (e, s1, s2) =
			let (_, i1, s1) = EqSet.trySubstM i msk s in
			match x with
			| None -> ("trySubstM did change s", s, s1)
			| Some x ->
				let c = Cons.mk (i1 + 1) (eq [1, x] 0) in
				let (_, s2) = EqSet.subst x i c s1 in
				let e =
					"still some " ^ (Basis.pr b x) ^ " in\n" ^ (EqSet.prExt b s1)
				in
				(e, s1, s2)
		in
		OUnit.assert_bool e (EqSet.isEq s1 s2)))
	in
	let tcs = [
		"x0", 0, Some x, mask [x], mk [
			y, eq [1, x; 1, y] 0;
			x, eq [1, x] 0 ];

		"no0", 0, None, mask [z], mk [
			x, eq [1, x] 0 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("var", OUnit.TestList (List.map chk_var tcs));
		OUnit.TestLabel ("nox", OUnit.TestList (List.map chk_nox tcs))
	]

(* EqSet.setupJoin *)
let setupJoin_ts =
	let (alpha, b1) = Basis.add b "alpha" in
	let chk_id dup (t, idOff, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let (_, _, s1) = EqSet.setupJoin dup b1 Rtree.Nil alpha idOff s in
		let sorted =
			let ids = List.map (fun c -> c.Cons.id) (EqSet.list s) in
			List.sort Pervasives.compare ids
		in
		let sorted1 =
			let ids1 = List.map (fun c -> c.Cons.id - idOff) (EqSet.list s1) in
			List.sort Pervasives.compare ids1
		in
		OUnit.assert_equal sorted sorted1))
	in
	let chk_def dup (t, i0, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let (b2, reloc, s1) = EqSet.setupJoin dup b1 Rtree.Nil alpha i0 s in
		let a =
			let chk (x1, _) (x2, _) =
				if dup then
					x1 = x2
				else
					Rtree.get None reloc x1 = Some x2
			in
			List.for_all2 chk s s1
		in
		let e =
			"wrong defined variables\n" ^
			"s:\n" ^ (EqSet.prExt b s) ^ "\n" ^
			"s1:\n" ^ (EqSet.prExt b2 s1) ^ "\n"
		in
		OUnit.assert_bool e a))
	in
	let chk_frag dup (t, i0, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let (_, _, s1) = EqSet.setupJoin dup b1 Rtree.Nil alpha i0 s in
		let a =
			let chk c =
				match c.Cons.f with
				| [] -> false
				| (id, coef)::[] -> id = c.Cons.id && Num.cmp Num.u coef = 0
				| _ -> false
			in
			List.for_all chk (EqSet.list s1)
		in
		let e = "bad fragments" in
		OUnit.assert_bool e a))
	in
	let tcs = [
		"one0", 10, mk [ x, eq [1, x] 1 ];
		"one1", 10, mk [ x, eq [1, x; 1, y] 1 ];
		"one2", 10, mk [ y, eq [1, x; 1, y] 1 ];
		"m0", 10, mk [
			x, eq [1, x; 1, y] 1;
			y, eq [2, y; 1, z] 1;
			z, eq [1, z; 2, t] 1 ]
	] in
	let chk_id = chk_id false and chk_idd = chk_id true
	and chk_def = chk_def false and chk_frag = chk_frag false
	and chk_defd = chk_def true and chk_fragd = chk_frag true
	in
	OUnit.TestList [
		OUnit.TestLabel ("id", OUnit.TestList (List.map chk_id tcs));
		OUnit.TestLabel ("idd", OUnit.TestList (List.map chk_idd tcs));
		OUnit.TestLabel ("def", OUnit.TestList (List.map chk_def tcs));
		OUnit.TestLabel ("defd", OUnit.TestList (List.map chk_defd tcs));
		OUnit.TestLabel ("frag", OUnit.TestList (List.map chk_frag tcs));
		OUnit.TestLabel ("fragd", OUnit.TestList (List.map chk_fragd tcs))
	]

(* EqSet.isIncl *)
let isIncl_ts =
	let chk_res (t, r, s1, s2) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let r1 =
			match EqSet.isIncl s1 s2 with
			| EqSet.NoIncl -> false
			| EqSet.Incl _ -> true
		in
		let e =
			(if r then "" else "un") ^ "expected inclusion" ^
			"\ns1:\n" ^ (EqSet.prExt b s1) ^ "\ns2:\n" ^ (EqSet.prExt b s2)
		in
		OUnit.assert_equal ~msg:e r r1))
	in
	let chk_cert (t, _, s1, s2) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		match EqSet.isIncl s1 s2 with
		| EqSet.NoIncl -> ()
		| EqSet.Incl w ->
			let c = Cert.mk w in
			let e =
				"can't check\n" ^ (Cert.pr c) ^
				"\ns1:\n" ^ (EqSet.prExt b s1) ^ "\ns2:\n" ^ (EqSet.prExt b s2)
			in
			let mklist s =
				List.map (fun c -> (c.Cons.id, c.Cons.c)) (EqSet.list s)
			in
			try
				OUnit.assert_bool e (Glue.check (mklist s1) (mklist s2) c)
			with
				Failure "Cert.combine" -> OUnit.assert_bool e false))
	in
	let mk l =
		let add (i, s) c =
			match EqSet.add s (Cons.mk i c) with
			| EqSet.Bottom _ -> failwith "EqSet_t.isIncl_ts"
			| EqSet.Ok (i1, s1) -> (i1, s1)
		in
		let (_, s) = List.fold_left add (0, EqSet.nil) l in
		s
	in
	let tcs = [
		"id0", true, mk [ eq [1, x] 0 ], mk [ eq [1, x] 0 ];
		"id1", true, mk [
			eq [1, y] 0;
			eq [1, x] 0
		], mk [ eq [1, x] 0 ];

		"bin0", true, mk [
			eq [1, y] 0;
			eq [1, x] 0
		], mk [ eq [1, x; 1, y] 0 ];

		"subst0", true, mk [
			eq [1, x] 0;
			eq [1, x; 1, y] 0
		], mk [ eq [1, y] 0 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("res", OUnit.TestList (List.map chk_res tcs));
		OUnit.TestLabel ("cert", OUnit.TestList (List.map chk_cert tcs))
	]

(* EqSet.rename *)
let rename_ts =
	let (x', b) = Basis.add b "x'" in
	let chkVar (t, fromX, toY, s0) =
		OUnit.TestLabel(t, OUnit.TestCase (fun () ->

		let defVars s = List.map (fun (v, _) -> v) s in
		let s = EqSet.rename s0 fromX toY in
		let e = (Basis.pr b fromX) ^ " remaining\n" ^ (EqSet.prExt b s) in
		OUnit.assert_bool e (List.for_all (fun v -> v <> fromX) (defVars s))))
	in
	let chkCoef (t, fromX, toY, s0) =
		OUnit.TestLabel(t, OUnit.TestCase (fun () ->

		let aX s =
			let coef c x = Vec.get c.Cons.c.Cstr.v x in
			List.map (fun (_, c) -> coef c fromX) s
		in
		let s = EqSet.rename s0 fromX toY in
		let e = (Basis.pr b fromX) ^ " remaining\n" ^ (EqSet.prExt b s) in
		OUnit.assert_bool e (List.for_all (fun a -> Num.cmpz a = 0) (aX s))))
	in
	let mk l =
		let add (i, s) c =
			match EqSet.add s (Cons.mk i c) with
			| EqSet.Bottom _ -> failwith "EqSet_t.isIncl_ts"
			| EqSet.Ok (i1, s1) -> (i1, s1)
		in
		let (_, s) = List.fold_left add (0, EqSet.nil) l in
		s
	in
	let tcs = [
		"one0", x', x, mk [ eq [1, x'] 0 ];
		"one1", x', x, mk [ eq [1, x'; 1, y] 0 ];
		"no0", x', x, mk [ eq [1, y] 0 ];
		"m0", x', x ,mk [
			eq [1, x'; 1, y] 1;
			eq [1, x'; 2, y] 2 ];
	] in
	OUnit.TestList [
		OUnit.TestLabel ("var", OUnit.TestList (List.map chkVar tcs));
		OUnit.TestLabel ("coef", OUnit.TestList (List.map chkCoef tcs))
	]

let ts = OUnit.TestList [
		OUnit.TestLabel ("filter", filter_ts);
		OUnit.TestLabel ("add", add_ts);
		OUnit.TestLabel ("pick", pick_ts);
		OUnit.TestLabel ("subst", subst_ts);
		OUnit.TestLabel ("tryDefs", tryDefs_ts);
		OUnit.TestLabel ("trySubstM", trySubstM_ts);
		OUnit.TestLabel ("setupJoin", setupJoin_ts);
		OUnit.TestLabel ("isIncl", isIncl_ts);
		OUnit.TestLabel ("rename", rename_ts)
	]
