(* IneqSet *)
let (x, b) = Basis.add Basis.nil "x"
let (y, b) = Basis.add b "y"
let (z, b) = Basis.add b "z"
let (t1, b) = Basis.add b "t1"
let (t2, b) = Basis.add b "t2"
let (t3, b) = Basis.add b "t3"

let mkc t v c =
	Cstr.mk t (List.map (fun (n, x) -> (Num.mk n, x)) v) (Num.mk c)

let eq = mkc Cmp.Eq and le = mkc Cmp.Le and lt = mkc Cmp.Lt

let mask l =
	List.fold_left (fun m x -> Rtree.set None m x (Some x)) Rtree.Nil l

let optxpr =
	function
	| None -> "None"
	| Some x -> "Some " ^ (Basis.pr b x)

let mkN = List.map (fun (i, n) -> (i, Num.mk n))

let propPr = function
	| IneqSet.Empty f -> Printf.sprintf "Empty: %s" (Cert.prF f)
	| IneqSet.Trivial -> "Trivial"
	| IneqSet.Implied (i, ce) -> Printf.sprintf "Implied %i: %s" i (Cert.prC ce)
	| IneqSet.Check c -> Printf.sprintf "Check: %s" (Cons.prExt b c)

let mkEqSet =
	let add s (i, c) =
		match EqSet.add s (Cons.mk i c) with
		| EqSet.Bottom _ -> failwith "IneqSet_t.mkEqSet"
		| EqSet.Ok (_, s1) -> s1
	in
	List.fold_left add EqSet.nil

let mkIneqSet es =
	let add s (i, c) =
		match IneqSet.add b es s (Cons.mk i c) with
		| IneqSet.MeetBot _ -> failwith "IneqSet_t.mkIneqSet: Bottom"
		| IneqSet.NewEqs _ -> failwith "IneqSet_t.mkIneqSet: NewEqs"
		| IneqSet.MeetOk (_, s1) -> s1
	in
	List.fold_left add IneqSet.nil

(* IneqSet.synIncl *)
let synInclCheck_ts =
	let chk (t, es, s, cIn, cOut) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let eq c1 c2 =
			match c1, c2 with
			| IneqSet.Check c1, IneqSet.Check c2 ->
				(c1.Cons.id = c2.Cons.id) &&
				Cons.isEq c1 c2 &&
				Cert.isEqF c1.Cons.f c2.Cons.f

			| _, _ -> false
		in
		let es1 = mkEqSet es in
		OUnit.assert_equal ~printer:propPr ~cmp:eq
			(IneqSet.Check cOut) (IneqSet.synIncl es1 (mkIneqSet es1 s) cIn)))
	in
	let c i c f = { Cons.id = i; Cons.c = c; Cons.f = Cert.mkF (mkN f) } in
	let tcs = [
		"subst0",
			[ 0, eq [1, y] 0 ], [],
			c 2 (le [1, x; 1, y] 0) [2, 1],
			c 2 (le [1, x] 0) [0, -1; 2, 1]
	] in
	OUnit.TestList (List.map chk tcs)

let synIncl_ts = OUnit.TestList [
		OUnit.TestLabel ("check", synInclCheck_ts)
	]

(* IneqSet.add *)
let add_ts =
	let chk_id (t, r, _, c, eqList, ineqList, _) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let es = mkEqSet eqList in
		let s = mkIneqSet es ineqList in
		match IneqSet.add b es s c with
		| IneqSet.MeetBot _ -> ()
		| IneqSet.MeetOk (id2, _) | IneqSet.NewEqs ({ Cons.id = id2; _ }, _, _) ->

		let (e, r1) =
			if r then
				("no increment", id2 = c.Cons.id + 1)
			else
				("increment", id2 = c.Cons.id)
		in
		OUnit.assert_bool e r1))
	in
	let chk_l (t, _, n, c, eqList, ineqList, _) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let es = mkEqSet eqList in
		let s = mkIneqSet es ineqList in
		let (a, e) =
			match IneqSet.add b es s c, n with
			| IneqSet.MeetBot f, None ->
				let clist = List.concat [IneqSet.list s; EqSet.list es] in
				let list = List.map (fun c -> (c.Cons.id, c.Cons.c)) (c::clist) in
				if Glue.check list [] (Cert.Empty f) then
					(true, "")
				else
					let err = Printf.sprintf "cannot check certificate\n%s%s%s\n%s\n"
						(EqSet.prExt b es) (IneqSet.prExt b s) (Cons.prExt b c) (Cert.prF f)
					in
					(false, err)

			| IneqSet.MeetBot _, Some _ -> (false, "unexpectedly empty")
			| IneqSet.MeetOk _, None | IneqSet.NewEqs _, None -> (false, "not empty")
			| IneqSet.MeetOk (_, s1), Some n | IneqSet.NewEqs (_, s1, _), Some n ->
				let n1 = List.length (IneqSet.list s1) in
				let e =
					"expected " ^ (string_of_int n) ^
					" but got " ^ (string_of_int n1)
				in
				(n = n1, e)
		in
		OUnit.assert_bool e a))
	in
	let chk_subst (t, _, n, c, eqList, ineqList, _) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let es = mkEqSet eqList in
		let s = mkIneqSet es ineqList in
		match n with
		| None -> ()
		| Some _ ->

		match IneqSet.add b es s c with
		| IneqSet.MeetBot _ -> failwith "IneqSet_t.chk_subst"
		| IneqSet.MeetOk (_, s1) | IneqSet.NewEqs (_, s1, _) ->

		let vars = List.map (fun (x, _) -> x) es in
		let chk s =
			let zcoef c x = Num.cmpz (Vec.get c.Cons.c.Cstr.v x) = 0 in
			List.for_all (fun c -> List.for_all (zcoef c) vars) s
		in
		if chk s then
			let e = "unprojected form\ns:\n" ^ (IneqSet.prExt b s1) in
			OUnit.assert_bool e (chk s1)
		else
			failwith "IneqSet_t.chk_subst"))
	in
	let chk_eqs (t, _, n, c, eqList, ineqList, newEqs) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let es = mkEqSet eqList in
		let s = mkIneqSet es ineqList in
		match n with
		| None -> ()
		| Some _ ->

		match IneqSet.add b es s c with
		| IneqSet.MeetBot _ -> failwith "IneqSet_t.chk_eqs"
		| IneqSet.MeetOk _ -> OUnit.assert_equal ~msg:"no new equalities found" newEqs []
		| IneqSet.NewEqs (_, _, eqs) ->
			let eqs = List.map Pervasives.fst eqs in
			let incl s1 s2 =
				List.for_all (fun c2 -> List.exists
					(fun c1 -> Cstr.isEqSyn c1.Cons.c c2.Cons.c) s1) s2
			in
			let e =
				(List.fold_left (fun s c -> s ^ c ^ "\n")
					"bad infered equalities:\n"
					(List.map (fun c -> Cons.prExt b c) eqs))
			in
			OUnit.assert_bool e (incl eqs newEqs && incl newEqs eqs)))
	in
	let c i c = Cons.mk i c in
	let tcs = [
		"nil0", true, Some 1, c 0 (le [1, x] 0), [], [], [];
		"new0", true, Some 2, c 1 (le [1, x] 0), [], [0, le [1, y] 0], [];
		"trivial0", false, Some 0, c 0 (le [] 0), [], [], [];
		"trivial1", false, Some 1, c 1 (le [] 0), [], [0, le [1, x] 0], [];
		"rem0", true, Some 1, c 1 (le [1, x] 0), [], [0, le [1, x] 1], [];
		"rem1", true, Some 2, c 2 (le [1, x] 0), [], [
			0, le [1, x; 1, y] 1;
			1, le [1, y] 1 ], [];

		"contrad0", false, None, c 0 (le [] (-1)), [], [], [];
		"contrad1", false, None, c 1 (le [] (-1)), [], [0, le [1, x] 1], [];
		"contrad2", false, None, c 1 (le [1, x] 1), [0, eq [1, x] 2], [], [];
		"contrad3", false, None, c 1 (le [1, x] 2), [], [0, lt [-1, x] (-2)], [];
		"contrad4", false, None, c 3 (le [1, x; 1, y] 2), [], [
			0, le [-1, x] (-3);
			1, le [-2, y] 1 ], [];
		"contrad5", false, None, c 3 (le [1, x; 1, z] 1), [0, eq [1, z] 0], [
			1, le [1, y] 1;
			2, le [-1, x; -1, y] (-3)], [];

		"incl0", false, Some 0, c 1 (le [1, x] 1), [0, eq [1, x] 0], [], [];
		"incl1", false, Some 0, c 3 (le [1, x; 1, y] 1), [
			0, eq [1, x] 1;
			1, eq [1, z] 1;
			2, eq [1, y; 2, z] 1
		], [], [];

		"incl2", false, Some 2, c 2 (le [1, x; 1, z] 1), [], [
			0, le [1, x; 2, y] 0;
			1, le [-2, y; 1, z] 1 ], [];

		"incl3", false, Some 1, c 1 (le [1, x] 0), [], [0, le [1, x] 0], [];
		"subst0", true, Some 2, c 2 (le [1, x; 1, y] 1), [
			0, eq [1, x; -1, y] 0 ], [
			1, le [1, z] 0
		], [];

		"eq0", false, Some 0, c 1 (le [1, x] 0), [], [0, le [-1, x] 0], [
			c 0 (eq [-1, x] 0);
			c 1 (eq [1, x] 0) ];

		"eq1", false, Some 0, c 1 (le [1, x; 1, y] 0), [], [
			0, le [-1, x; -1, y] 0
		 ], [
			c 0 (eq [-1, x; -1, y] 0);
			c 1 (eq [1, x; 1, y] 0) ];

		"eq2", false, Some 1, c 2 (le [1, x; 1, y] 0), [], [
			0, le [1, z] 1;
			1, le [-1, x; -1, y] 0
		 ], [
			c 0 (eq [-1, x; -1, y] 0);
			c 1 (eq [1, x; 1, y] 0) ];

		"eqs0", false, Some 0, c 2 (le [-1, x; 1, z] 0), [], [
			0, le [1, x; -1, y] 0;
			1, le [1, y; -1, z] 0
		 ], [
			c 0 (eq [1, x; -1, y] 0);
			c 1 (eq [1, y; -1, z] 0);
			c 2 (eq [-1, x; 1, z] 0) ];

		"eqs1", false, Some 0, c 4 (le [-1, x; 1, z] 0), [], [
			0, le [1, x; -1, y] 0;
			1, le [1, y; -1, z] 0;
			2, le [1, t1; 1, x; -1, z; -1, t2] 0;
			3, le [1, t2; -1, t1] 0
		], [
			c 0 (eq [1, x; -1, y] 0);
			c 1 (eq [1, y; -1, z] 0);
			c 2 (eq [1, t1; 1, x; -1, z; -1, t2] 0);
			c 3 (eq [1, t2; -1, t1] 0);
			c 4 (eq [-1, x; 1, z] 0) ];

		"eqs2", false, Some 1, c 3 (le [-1, x; 1, z] 0), [], [
			0, le [1, x; -1, y] 0;
			1, le [1, y; -1, z] 0;
			2, le [1, t3] 0
		], [
			c 0 (eq [1, x; -1, y] 0);
			c 1 (eq [1, y; -1, z] 0);
			c 2 (eq [-1, x; 1, z] 0) ];

		"eqs3", false, Some 1, c 5 (le [-1, x; 1, z] 0), [], [
			0, le [1, x; -1, y] 0;
			1, le [1, y; -1, z] 0;
			2, le [1, t1; 1, x; -1, z; -1, t2] 0;
			3, le [1, t2; -1, t1] 0;
			4, le [1, t3] 0
		], [
			c 0 (eq [1, x; -1, y] 0);
			c 1 (eq [1, y; -1, z] 0);
			c 2 (eq [-1, x; 1, z] 0);
			c 3 (eq [1, t2; -1, t1] 0);
			c 4 (eq [1, t1; 1, x; -1, z; -1, t2] 0) ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("id", OUnit.TestList (List.map chk_id tcs));
		OUnit.TestLabel ("l", OUnit.TestList (List.map chk_l tcs));
		OUnit.TestLabel ("subst", OUnit.TestList (List.map chk_subst tcs));
		OUnit.TestLabel ("eqs", OUnit.TestList (List.map chk_eqs tcs))
	]

(* IneqSet.pick *)
let pick_ts =
	let chk (t, msk, x, l) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let s = mkIneqSet EqSet.nil l in
		let x1 = IneqSet.pick msk s in
		OUnit.assert_equal ~printer:optxpr x x1))
	in
	let tcs = [
		"nil0", mask [], None, [];
		"nil1", mask [x], None, [];
		"nil2", mask [y], None, [];
		"no0", mask [], None, [0, le [1, x] 0];
		"no1", mask [y], None, [0, le [1, x] 0];
		"no2", mask [x], None, [0, le [1, z] 0];
		"one0", mask [x], Some x, [0, le [1, x] 0];
		"one1", mask [x], Some x, [
			0, le [1, x] 0;
			1, le [1, y] 0 ];

		"one2", mask [x; y], Some x, [
			0, le [1, x] 0;
			1, le [1, x; 1, y] 0 ];

		"m0", mask [x; y], Some y, [
			0, le [1, x; 2, z] 0;
			1, le [1, x; 1, z] 0;
			2, le [-1, x; 1, y] 0;
			3, le [2, y; 1, z] 0 ]
	] in
	OUnit.TestList (List.map chk tcs)

(* IneqSet.subst *)
let subst_ts =
	let chk_id (t, x, i, i1, e, s, _) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let s1 = mkIneqSet EqSet.nil s in
		let (i2, _) = IneqSet.subst b EqSet.nil x i e s1 in
		OUnit.assert_equal ~printer:string_of_int i1 i2))
	in
	let chk_res (t, x, i, _, e, s, s1) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let s2 = mkIneqSet EqSet.nil s and s3 = mkIneqSet EqSet.nil s1 in
		let (_, s4) = IneqSet.subst b EqSet.nil x i e s2 in
		OUnit.assert_equal ~printer:(IneqSet.prExt b) ~cmp:IneqSet.isEq s3 s4))
	in
	let c i c = Cons.mk i c in
	let tcs = [
		"nil0", x, 1, 1, c 0 (eq [1, x] 0), [], [];
		"other0", x, 1, 2, c 0 (eq [1, x] 0), [
			1, le [1, y] 0
		], [
			1, le [1, y] 0 ];

		"other1", y, 1, 3, c 0 (eq [1, y] 0), [
			1, le [1, x] 0;
			2, le [1, x; 1, z] 0
		], [
			1, le [1, x] 0;
			2, le [1, x; 1, z] 0 ];

		"one0", x, 1, 2, c 0 (eq [1, x] 0), [
			1, le [1, x; 1, y] 0;
		], [
			1, le [1, y] 0 ];

		"mixed0", x, 1, 3, c 0 (eq [1, x] 0), [
			1, le [1, x; 1, y] 0;
			2, le [1, z] 0
		], [
			1, le [1, y] 0;
			2, le [1, z] 0 ];

		"simpl0", x, 1, 2, c 0 (eq [1, x] 0), [
			1, le [1, x; 1, y] 0;
			2, le [1, y] 0
		], [
			1, le [1, y] 0 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("id", OUnit.TestList (List.map chk_id tcs));
		OUnit.TestLabel ("res", OUnit.TestList (List.map chk_res tcs))
	]

(* IneqSet.fmElim *)
let fmElim_ts =
	let chk_id (t, x, i, i1, s, _) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let s1 = mkIneqSet EqSet.nil s in
		let (i2, _) = IneqSet.fmElim b EqSet.nil x i s1 in
		OUnit.assert_equal ~printer:string_of_int i1 i2))
	in
	let chk_res (t, x, i, _, s, s1) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let s2 = mkIneqSet EqSet.nil s and s3 = mkIneqSet EqSet.nil s1 in
		let (_, s4) = IneqSet.fmElim b EqSet.nil x i s2 in
		OUnit.assert_equal ~printer:(IneqSet.prExt b) ~cmp:IneqSet.isEq s3 s4))
	in
	let tcs = [
		"nil0", x, 1, 1, [], [];
		"other0", x, 1, 2, [
			1, le [1, y] 0
		], [
			1, le [1, y] 0 ];

		"plus0", x, 1, 2, [
			1, le [1, x] 0;
			2, le [1, y] 0
		], [
			1, le [1, y] 0 ];

		"plus1", x, 1, 2, [
			1, le [1, x] 0;
			2, le [1, x; 1, y] 0;
			3, le [1, y] 0
		], [
			2, le [1, y] 0 ];

		"plus2", x, 1, 1, [
			1, le [1, x] 0
		], [];

		"neg0", x, 1, 2, [
			1, le [-1, x] 0;
			2, le [1, y] 0
		], [
			1, le [1, y] 0 ];

		"m0", x, 1, 2, [
			1, le [1, x; 1, y] 0;
			2, le [-1, x] 0
		], [
			1, le [1, y] 0
		];

		"m1", x, 1, 1, [
			1, le [1, x] 1;
			2, le [-1, x] 0
		], [];

		"m2", x, 1, 3, [
			1, le [1, x; 1, y] 0;
			2, le [1, x; 1, z] 0;
			3, le [-1, x] 0
		], [
			1, le [1, y] 0;
			2, le [1, z] 0 ];

		"mixed0", x, 1, 3, [
			1, le [-1, x] 0;
			2, le [1, x; 1, y] 0;
			3, le [1, z] 0
		], [
			1, le [1, y] 0;
			2, le [1, z] 0 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("id", OUnit.TestList (List.map chk_id tcs));
		OUnit.TestLabel ("res", OUnit.TestList (List.map chk_res tcs))
	]

(* IneqSet.fmElimM *)
let fmElimM_ts =
	let chk_id (t, i, i1, msk, cList, _) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let s = mkIneqSet EqSet.nil cList in
		let (i2, _) = IneqSet.fmElimM b EqSet.nil i msk s in
		OUnit.assert_equal ~printer:string_of_int i1 i2))
	in
	let chk_res (t, i, _, msk, cList, cList1) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let s = mkIneqSet EqSet.nil cList and s1 = mkIneqSet EqSet.nil cList1 in
		let (_, s2) = IneqSet.fmElimM b EqSet.nil i msk s in
		OUnit.assert_equal ~printer:(IneqSet.prExt b) ~cmp:IneqSet.isEq s1 s2))
	in
	let tcs = [
		"nil0", 0, 0, mask [], [], [];
		"nil1", 0, 0, mask [x], [], [];
		"nil2", 0, 1, mask [], [
			1, le [1, x] 0
		], [
			1, le [1, x] 0 ];

		"two0", 0, 0, mask [x; y], [
			1, le [1, x] 0
		], [];

		"two1", 0, 0, mask [x; y], [
			1, le [1, x] 0;
			2, le [-1, x; 1, y] 0
		], [];

		"two2", 0, 1, mask [x; y], [
			1, le [1, x] 0;
			2, le [-1, x; 1, y] 0;
			3, le [-1, y; 1, z] 0
		], [
			1, le [1, z] 0 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("id", OUnit.TestList (List.map chk_id tcs));
		OUnit.TestLabel ("res", OUnit.TestList (List.map chk_res tcs))
	]

(* IneqSet.isIncl *)
let isIncl_ts =
	let chkRes (t, r, el, l1, l2) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let es = mkEqSet el in
		let s1 = mkIneqSet es l1 and s2 = mkIneqSet EqSet.nil l2 in
		let (a, s) =
			match IneqSet.isIncl b es s1 s2 with
			| IneqSet.NoIncl -> (not r, "expected inclusion")
			| IneqSet.Incl _ -> (r, "unexpected inclusion")
		and e =
			"\nes:\n" ^ (EqSet.prExt b es) ^
			"\ns1:\n" ^ (IneqSet.prExt b s1) ^
			"\ns2:\n" ^ (IneqSet.prExt b s2)
		in
		OUnit.assert_bool (s ^ e) a))
	in
	let chkCert (t, _, el, l1, l2) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let es = mkEqSet el in
		let s1 = mkIneqSet es l1 and s2 = mkIneqSet EqSet.nil l2 in
		match IneqSet.isIncl b es s1 s2 with
		| IneqSet.NoIncl -> ()
		| IneqSet.Incl w ->
			let c = Cert.mk w in
			let e =
				"can't check certificate\n" ^ (Cert.pr c) ^
				"\nes:\n" ^ (EqSet.prExt b es) ^
				"\ns1:\n" ^ (IneqSet.prExt b s1) ^
				"\ns2:\n" ^ (IneqSet.prExt b s2)
			in
			let mkl s = List.map (fun c -> (c.Cons.id, c.Cons.c)) s in
			try
				let s = mkl (List.append (IneqSet.list s1) (EqSet.list es)) in
				OUnit.assert_bool e (Glue.check s (mkl (IneqSet.list s2)) c)
			with
				Failure "Cert.combine" -> OUnit.assert_bool e false))
	in
	let tcs = [
		"simpl0", true, [], [
			0, le [1, x] 1
		], [
			0, le [1, x] 2 ];

		"simpl1", false, [], [
			0, le [1, x] 2
		], [
			0, le [1, x] 1 ];

		"substSimpl0", true, [0, eq [1, x] 0], [
			1, le [1, x; 1, y] 1
		], [
			0, le [1, y] 2 ];

		"substSimpl1", true, [0, eq [1, x] 0], [
			1, le [1, y] 1
		], [
			0, le [1, x; 1, y] 1 ];

		"substComb0", true, [0, eq [1, x] 0], [
			1, le [1, x; 1, y] 1;
			2, le [1, z] 1
		], [
			0, le [1, y; 1, z] 2 ];

		"substTriv0", true, [0, eq [1, x] 0], [], [
			0, le [-1, x] 0 ];

		"substTriv1", true, [0, eq [1, x] 0], [
			1, le [1, y] 0
		], [
			0, le [1, y] 0;
			1, le [-1, x] 0 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("res", OUnit.TestList (List.map chkRes tcs));
		OUnit.TestLabel ("cert", OUnit.TestList (List.map chkCert tcs))
	]

(* IneqSet.fromUnsat *)
let fromUnsatKo_ts =
	let chk (t, i, w) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		OUnit.assert_raises (Invalid_argument "IneqSet.fromUnsat")
			(fun () -> IneqSet.fromUnsat (mkN w) i)))
	in
	let tcs = [
		"empty0", 1, [];
		"noI0", 1, [0, 1];
		"noI1", 1, [0, 1; 2, 3];
		"many0", 0, [0, 1; 0, 2];
		"many1", 0, [1, 2; 0, 1; 2, -5; 0, 2];
		"bad0", 1, [0, 1; 1, 0; 2, 3]
	] in
	List.map chk tcs

let fromUnsatOk_ts =
	let chk (t, i, w, f) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		OUnit.assert_equal ~printer:Cert.prF ~cmp:Cert.isEqF
			(Cert.mkF (mkN f)) (IneqSet.fromUnsat (mkN w) i)))
	in
	let tcs = [
		"unit0", 1, [1, 1; 2, 1], [2, 1];

		(* normalization *)
		"unit1", 1, [1, 2; 2, 4], [2, 2];
		"unit2", 1, [1, -1; 3, 2], [3, 2]; (* XXX: Is this really what we want? *)

		(* order *)
		"unit3", 1, [2, 3; 1, 1], [2, 3];
		"unit4", 2, [4, 2; 2, 2], [4, 1];

		"m0", 2, [1, 2; 4, 2; 2, 1], [1, 2; 4, 2];
		"m1", 2, [1, 2; 4, 2; 2, 2], [1, 1; 4, 1]

	] in
	List.map chk tcs

let fromUnsat_ts = OUnit.TestList [
		OUnit.TestLabel ("ko", OUnit.TestList fromUnsatKo_ts);
		OUnit.TestLabel ("ok", OUnit.TestList fromUnsatOk_ts)
	]

let ts = OUnit.TestList [
		OUnit.TestLabel ("synIncl", synIncl_ts);
		OUnit.TestLabel ("add", add_ts);
		OUnit.TestLabel ("pick", pick_ts);
		OUnit.TestLabel ("subst", subst_ts);
		OUnit.TestLabel ("fmElim", fmElim_ts);
		OUnit.TestLabel ("fmElimM", fmElimM_ts);
		OUnit.TestLabel ("isIncl", isIncl_ts);
		OUnit.TestLabel ("fromUnsat", fromUnsat_ts)
	]
