let (x, b) = Basis.add Basis.nil "x"
let (y, b) = Basis.add b "y"
let (z, b) = Basis.add b "z"
let (a, b) = Basis.add b "a"
let (c, b) = Basis.add b "c"
let (d, b) = Basis.add b "d"
let (e, b) = Basis.add b "e"

let mkc t v c =
	Cstr.mk t (List.map (function (i, x) -> (Num.mk i, x)) v) (Num.mk c)

let eq = mkc Cmp.Eq and le = mkc Cmp.Le and lt = mkc Cmp.Lt

let p (l: Cstr.t list): Pol.t =
	match Pol.mk b l with
	| None -> failwith "Pol.mk returned None"
	| Some poly -> poly

(* Pol.add *)
let add_ts =
	let chkBot (t, empty, c, p) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		match Pol.add p c with
		| (Some _, _) -> OUnit.assert_bool "should be bottom" (not empty)
		| (None, _ ) -> OUnit.assert_bool "should not be bottom" empty))
	in
	let chkCert (t, _, c, p) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let (p1, cert) = Pol.add p c in
		let prOpt = function
			| None -> "None"
			| Some p -> Printf.sprintf "Some p1:\n%s" (Pol.prExt p)
		in
		let msg = Printf.sprintf
			"can't check certificate for p1 = Pol.add p c\np:\n%s\nc:%s\n\np1:%s\ncert:\n%s\n"
				(Pol.prExt p) (Cons.prExt b (Cons.mk p.Pol.nxt c)) (prOpt p1) (Cert.pr cert)
		in
		OUnit.assert_bool msg (Pol.checkMeet p c p1 cert)))
	in
	let chk_subst (t, isEmpty, c, p) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		match Pol.add p c, isEmpty with
		| (None, _), true | (None, _), false | (Some _, _), true -> ()
		| (Some p1, _), false ->

		let defs = List.map (fun (x, _) -> x) p1.Pol.eqs in
		let chk1 c = List.for_all
			(fun x -> Num.cmpz (Vec.get c.Cons.c.Cstr.v x) = 0) defs
		in
		let err = "defined variables in inequality set\n" ^ (Pol.prExt p1) in
		OUnit.assert_bool err (List.for_all chk1 p1.Pol.ineqs)))
	in
	let tcs = [
		"noEqs0", false, le [1, x] 0, p [
			le [1, y] 0 ];

		"eqs0", false, le [1, x; 1, y] 0, p [
			eq [1, x; -1, y] 0 ];

		"eqs1", false, le [1, x; 1, y] 0, p [
			eq [1, x; -1, z] 0];

		"eqs2", false, eq [1, x; 1, y] 1, p [
			eq [1, x] 0];

		"eqs3", false, eq [1, x; 1, y] 1, p [
			eq [1, x] 0;
			le [1, x; 1, y; 1, z] 0];

		"newEq0", false, eq [1, x] 0, p [
			le [1, x; 1, y] 2 ];

		"implicitEq0", false, le [1, x] 0, p [
			le [1, y] 0;
			le [-1, x] 0 ];

		"implicitEqs0", false, le [1, z; -1, x] 0, p [
			le [1, y; -1, z] 0;
			le [1, x; -1, y] 0 ];

		"implicitEqs1", false, le [1, z; -1, x] 0, p [
			le [1, y; -1, z] 0;
			le [1, x; -1, y] 0;
			le [1, x; 1, a] 1 ];

		(* substitutions in the building blocks of an implicit equality *)
		"implicitEqs2", false, le [1, x; 1, y] 0, p [
			eq [1, y] 0;
			le [-1, x] 0 ];

		"emptyEq0", true, eq [-1, z] 2, p [
			eq [-1, x] 2;
			eq [-1, y] 0;
			le [-1, z] (-6) ];

		"emptyLe0", true, le [1, x] 0, p [
			le [-1, x] (-1) ];

		"emptyLe1", true, le [1, x; 1, y] 0, p [
			le [-1, x] (-1);
			eq [1, y] 0 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("bot", OUnit.TestList (List.map chkBot tcs));
		OUnit.TestLabel ("cert", OUnit.TestList (List.map chkCert tcs));
		OUnit.TestLabel ("subst", OUnit.TestList (List.map chk_subst tcs))
	]

(* Pol.isIncl *)
let isIncl_ts =
	let chk_res (t, r, p1, p2) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let r1 =
			match Pol.isIncl p1 p2 with
			| Pol.NoIncl -> false
			| Pol.Incl _ -> true
		in
		let err =
			(if r then "expected inclusion" else "unexpected inclusion") ^
			"\np1:\n" ^ (Pol.prExt p1) ^ "\np2:\n" ^ (Pol.prExt p2)
		in
		OUnit.assert_equal ~msg:err r r1))
	in
	let chk_cert (t, _, p1, p2) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		match Pol.isIncl p1 p2 with
		| Pol.NoIncl -> ()
		| Pol.Incl c ->
			let estr = "cannot check certificate\n" ^ (Cert.pr c) in
			try
				OUnit.assert_bool estr (Pol.check p1 p2 c)
			with
				Failure "Cert.combine" -> OUnit.assert_bool estr false))
	in
	let tcs = [
		"simple0", true, p [
			le [1, x; 1, y] 2
		], p [
			le [1, x; 1, y] 3 ];

		"simple1", true, p [
			le [1, x] 2
		], p [
			le [1, x] 3 ];

		"simple2", true, p [
			le [1, x] 2
		], p [
			le [2, x] 5 ];

		"eq0", true, p [
			le [1, x] 1;
			le [-1, x] (-1)
		], p [
			eq [1, x] 1 ];

		"fail0", false, p [
			le [1, x] 2
		], p [
			le [1, x] 1 ];

		"eq1", true, p [
			eq [1, x] 0
		], p [
			eq [1, x] 0 ];

		"eq2", true, p [
			eq [-1, x] 0
		], p [
			eq [-1, x] 0 ];

		"decompress", false, p [
			le [-1, y] (-1);
			le [-1, x; 16, y] 15;
			le [-1, x] (-17)
		], p [
			le [1, x; -16, y] 17 ];

		"sylvain", false, p [
			le [2, x] 3
		], p [
			le [1, x] 1 ];

		"splx0", true, p [
			le [1, x] 2;
			le [2, y] 1;
		], p [
			le [1, x; 2, y] 3 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("res", OUnit.TestList (List.map chk_res tcs));
		OUnit.TestLabel ("cert", OUnit.TestList (List.map chk_cert tcs))
	]

(* Pol.project *)
let project_ts =
	let chk_res (t, p, r) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let (p1, _) = Pol.project p x in
		let estr =
			"projecting x from polyhedron P:\n" ^ (Pol.prExt p) ^
			"\ngives polyhedron P1:\n" ^ (Pol.prExt p1) ^
			"\nwhich is not syntactically equal to the expected R:\n" ^
				(Pol.prExt r)
		in
		OUnit.assert_bool estr (Pol.isEqSyn p1 r)))
	in
	let chk_cert (t, p, _) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let (p1, c) = Pol.project p x in
		let estr =
			"cannot check inclusion of polyhedron P:\n" ^ (Pol.prExt p) ^
			"\nin P1:\n" ^ (Pol.prExt p1) ^
			"\nin using the following certificate:\n" ^ (Cert.pr c)
		in
		OUnit.assert_bool estr (Pol.check p p1 c)))
	in
	let tcs = [
		"empty", p [], p [];
		"no_x0", p [eq [1, y] 0], p [eq [1, y] 0];
		"no_x1", p [eq [1, y] 0; eq [2, y; 1, z] 0],
			p [eq [1, y] 0; eq [2, y; 1, z] 0];
		"only_x", p [eq [1, x] 0], p [];
		"simple0", p [le [1, x; 2, y] 1; le [1, y] 2], p [le [1, y] 2];
		"subst0", p [eq [1, x; -1, y] 0; le [1, x] 1], p [le [1, y] 1];
		"fme0", p [le [1, x] 0; le [-1, x; 1, y] 1], p [le [1, y] 1];
		"trivial0", p [le [1, x; 1, y] 0; le [-1, x; -1, y] 1], p [];
	] in
	OUnit.TestList [
		OUnit.TestLabel ("res", OUnit.TestList (List.map chk_res tcs));
		OUnit.TestLabel ("cert", OUnit.TestList (List.map chk_cert tcs))
	]

(* Pol.projectM *)
let projectM_ts =
	let chk_res (t, l, p, r) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let (p1, _) = Pol.projectM p l in
		let estr =
			"projecting" ^
			(List.fold_left (fun s x -> s ^ " " ^ (Basis.pr b x)) "" l) ^
 			" from polyhedron P:\n" ^ (Pol.prExt p) ^
			"\ngives polyhedron P1:\n" ^ (Pol.prExt p1) ^
			"\nwhich is not syntactically equal to the expected R:\n" ^
				(Pol.prExt r)
		in
		OUnit.assert_bool estr (Pol.isEqSyn p1 r)))
	in
	let chk_cert (t, l, p, _) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let (p1, c) = Pol.projectM p l in
		let estr =
			"cannot check inclusion of polyhedron P:\n" ^ (Pol.prExt p) ^
			"\nin P1:\n" ^ (Pol.prExt p1) ^
			"\nin using the following certificate:\n" ^ (Cert.pr c)
		in
		OUnit.assert_bool estr (Pol.check p p1 c)))
	in
	let tcs = [
		"empty", [x], p [], p [];
		"no_x0", [x], p [
			eq [1, y] 0
		], p [
			eq [1, y] 0 ];

		"no_x1", [x], p [
			eq [1, y] 0;
			eq [2, y; 1, z] 0
		], p [
			eq [1, y] 0;
			eq [2, y; 1, z] 0 ];

		"only_x", [x], p [
			eq [1, x] 0
		], p [];

		"simple0", [x], p [
			le [1, x; 2, y] 1;
			le [1, y] 2
		], p [
			le [1, y] 2 ];

		"subst0", [x], p [
			eq [1, x; -1, y] 0;
			le [1, x] 1
		], p [
			le [1, y] 1 ];

		"fme0", [x], p [
			le [1, x] 0;
			le [-1, x; 1, y] 1
		], p [
			le [1, y] 1 ];

		"trivial0", [x], p [
			le [1, x; 1, y] 0;
			le [-1, x; -1, y] 1
		], p [];

		"le0", [x; z], p [
			le [1, x; 2, y; -3, z] 0;
			le [-2, x; -1, z] 0;
			le [3, y; 1, z] 0
		], p [
			le [1, y] 0 ];

		"eq0", [x; z], p [
			le [1, x; 2, y; -3, z] 0;
			le [-2, x; -1, z] 0;
			eq [1, y; -1, z] 0
		], p [
			le [-1, y] 0 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("res", OUnit.TestList (List.map chk_res tcs));
		OUnit.TestLabel ("cert", OUnit.TestList (List.map chk_cert tcs))
	]

(* Pol.setupJoin *)
let setupJoin_ts =
	let (x, b) = Basis.add Basis.nil "x" in
	let (y, b) = Basis.add b "y" in
	let (alpha, b1) = Basis.add b "alpha" in (* order of addition matters. *)
	let (x1, b1) = Basis.add b1 "x'1" in
	let (y1, b1) = Basis.add b1 "y'1" in
	let pmk b l =
		match Pol.mk b l with
		| None -> failwith "Pol_t.setupJoin_ts"
		| Some p -> p
	in
	let p = pmk b and p1 = pmk b1 in
	let chk_pol (t, _, p1, p2, p0) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let (_, p, _) = Pol.setupJoin p1 p2 in
		let estr =
			"expected p:\n" ^ (Pol.prExt p0) ^
			"\n\nactual p:\n" ^ (Pol.prExt p)
		in
		OUnit.assert_bool estr (Pol.isEqSyn p0 p)))
	in
	let chk_vars (t, l0, p1, p2, _) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let (_, p, l) = Pol.setupJoin p1 p2 in
		let incl l1 l2 =
			let chk x y =
				x = y && Basis.pr b1 x = Basis.pr p.Pol.b x
			in
			List.for_all (fun x -> List.exists (chk x) l2) l1
		in
		OUnit.assert_bool "bad var list" (incl l l0 && incl l0 l)))
	in
	let chk_frag (t, _, p1, p2, _) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let (_, p, _) = Pol.setupJoin p1 p2 in
		let initP = { p with
			Pol.eqs = EqSet.initc p.Pol.eqs;
			Pol.ineqs = IneqSet.initc p.Pol.ineqs }
		in
		let chk =
			List.for_all2 (fun c1 c2 ->
				List.for_all2 (fun (x1, n1) (x2, n2) ->
					x1 = x2 && Num.cmp n1 n2 = 0) c1.Cons.f c2.Cons.f)
		in
		let e =
			"not equal\np:\n" ^ (Pol.prExt p) ^ "\ninitP:\n" ^ (Pol.prExt initP)
		in
		OUnit.assert_bool e
			(chk (EqSet.list p.Pol.eqs) (EqSet.list initP.Pol.eqs) &&
			chk (IneqSet.list p.Pol.ineqs) (IneqSet.list initP.Pol.ineqs))))
	in
	let tcs = [
		"one", [alpha; x1; y1], p [
			eq [1, x] 1;
			le [-1, y] 1
		], p [
			eq [1, x] 1;
			le [1, y] 2
		], p1 [
			eq [1, x; -1, x1; 1, alpha] 1;
			eq [1, x1; -1, alpha] 0;
			le [-1, y1; -1, alpha] 0;
			le [1, y; -1, y1; 2, alpha] 2;
			le [-1, alpha] 0;
			le [1, alpha] 1 ];

		"all", [alpha; x1; y1], p [
			eq [1, x] 1
		], p [
			eq [1, y] 1
		], p1 [
			eq [1, y; -1, y1; 1, alpha] 1;
			eq [1, x1; -1, alpha] 0;
			le [-1, alpha] 0;
			le [1, alpha] 1 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("pol", OUnit.TestList (List.map chk_pol tcs));
		OUnit.TestLabel ("vars", OUnit.TestList (List.map chk_vars tcs));
		OUnit.TestLabel ("frag", OUnit.TestList (List.map chk_frag tcs))
	]

(* Pol.join *)
let join_ts =
	let chk_incl (t, p1, p2) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let (p,_,_) = Pol.join p1 p2 in
		let pstr = "p:\n" ^ (Pol.prExt p) in
		let p1str = "p1 not in p.\np1:\n" ^ (Pol.prExt p1) ^ "\n" in
		let p2str = "p2 not in p.\np2:\n" ^ (Pol.prExt p2) ^ "\n"
		in
		let check pn =
			match Pol.isIncl pn p with
			| Pol.NoIncl -> false
			| Pol.Incl _ -> true
		in
		OUnit.assert_bool (p1str ^ pstr) (check p1);
		OUnit.assert_bool (p2str ^ pstr) (check p2)))
	in
	let chk_cert1 (t, p1, p2) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let (_, pb, vars) = Pol.setupJoin p1 p2
		and (p, c, _) = Pol.join p1 p2 in
		let (_, rawCert) = Pol.projectM pb vars in
		let diag =
			"\n" ^ (Cert.pr c) ^
			"\np1:\n" ^ (Pol.prExt p1) ^ "\np:\n" ^ (Pol.prExt p) ^
			"\nraw certificate (" ^
			"p1.nxt: " ^ (string_of_int p1.Pol.nxt) ^
			", p2.nxt: " ^ (string_of_int p2.Pol.nxt) ^
			"):\n" ^ (Cert.pr rawCert) ^
			"\npb:\n" ^ (Pol.prExt pb)
		in
		let (a, err) =
			try
				(Pol.check p1 p c, "bad cert")
			with
				Failure "Cert.combine" -> (false, "Failure: Cert.Combine")
		in
		OUnit.assert_bool (err ^ diag) a))
	in
	let chk_cert2 (t, p1, p2) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let (_, pb, vars) = Pol.setupJoin p1 p2
		and (p, _, c) = Pol.join p1 p2 in
		let (_, rawCert) = Pol.projectM pb vars in
		let diag =
			"\n" ^ (Cert.pr c) ^
			"\np2:\n" ^ (Pol.prExt p2) ^ "\np:\n" ^ (Pol.prExt p) ^
			"\nraw certificate (" ^
			"p1.nxt: " ^ (string_of_int p1.Pol.nxt) ^
			", p2.nxt: " ^ (string_of_int p2.Pol.nxt) ^
			"):\n" ^ (Cert.pr rawCert) ^
			"\npb:\n" ^ (Pol.prExt pb)
		in
		let (a, err) =
			try
				(Pol.check p2 p c, "bad cert")
			with
				Failure "Cert.combine" -> (false, "Failure: Cert.Combine")
		in
		OUnit.assert_bool (err ^ diag) a))
	in
	let tcs = [
		"one", p [
			eq [1, x] 1;
			le [-1, y] 1
		], p [
			eq [1, x] 1;
			le [1, y] 2 ];

		"all",p [
			eq [1, x] 1
		], p [
			eq [1, y] 1 ];

		"chull969", p [
			eq [1, c; -1, d] (-4080);
			eq [-1, e] (-15);
			le [1, z] 6;
			le [1, c] 15;
			le [-1, a] (-256);
			le [-1, z] (-2);
			le [-1, y] (-1);
			le [-1, x; 1, z] 0
		], p [
			le [-16, e; 1, d] 3855 ];

		"sylvain", p [
			eq [-1, x; 1, y] 5;
			le [1, x] 5;
			le [-1, x] (-1)
		], p [
			eq [1, x; -1, y] 5;
			le [1, x] 10;
			le [-1, x] (-6) ];

		"jdphuff_chull777", p [
			eq [-1, x] 0;
			le [-1, y] 0;
			le [-1, z; 1, a] 1
		], p [
			eq [-1, x] 0;
			le [-1, y] 0;
			eq [-1, z; 1, a] 1 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("incl", OUnit.TestList (List.map chk_incl tcs));
		OUnit.TestLabel ("cert1", OUnit.TestList (List.map chk_cert1 tcs));
		OUnit.TestLabel ("cert2", OUnit.TestList (List.map chk_cert2 tcs))
	]

(* Pol.assign *)
let assign_ts =
	let chk (t, x, v, n, p1, p) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let p2 = Pol.assign p1 x v n in
		let estr =
			"result not syntactically equal to expected" ^
			"\nresult:\n" ^ (Pol.prExt p2) ^
			"\nexpected:\n" ^ (Pol.prExt p)
		in
		OUnit.assert_bool estr (Pol.isEqSyn p p2))) in
	let tcs = [
		"no_x0", x, Vec.nil, Num.u, p [], p [eq [1, x] 1];
		"no_x1", x, Vec.mk [Num.u, y], Num.z, p [],
			p [eq [1, x; -1, y] 0];
		"no_x2", x, Vec.mk [Num.u, y], Num.z, p [le [1, y] 1],
			p [eq [1, x; -1, y] 0; le [1, y] 1];
		"x0_proj", x, Vec.nil, Num.u, p [le [1, x] 2], p [eq [1, x] 1];
		"x0", x, Vec.mk [Num.u, x], Num.u, p [eq [1, x] 1],
			p [eq [-1, x] (-2)];

		"x1", x, Vec.mk [Num.u, x], Num.u, p [le [1, x] 1],
			p [le [1, x] 2]
	] in
		OUnit.TestList (List.map chk tcs)

(* Pol.assignM *)
let assignM_ts =
	let chk (t, l, p1, p) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let p2 = Pol.assignM p1 l in
		let err =
			"result not syntactically equal to expected" ^
			"\nresult:\n" ^ (Pol.prExt p2) ^
			"\nexpected:\n" ^ (Pol.prExt p)
		in
		OUnit.assert_bool err (Pol.isEqSyn p p2)))
	in
	let tcs = [
		"noX0", [{ Pol.var = x; Pol.lin = []; Pol.cst = Num.u }], p [],
			p [ eq [1, x] 1 ];

		"noX1", [{ Pol.var = x; Pol.lin = [Num.u, y]; Pol.cst = Num.z }],
			p [], p [ eq [1, x; -1, y] 0 ];

		"noX2", [{ Pol.var = x; Pol.lin = [Num.u, y]; Pol.cst = Num.z }],
			p [
				le [1, y] 1
			], p [
				eq [1, x; -1, y] 0;
				le [1, y] 1 ];

		"xProj0", [{ Pol.var = x; Pol.lin = []; Pol.cst = Num.u }],
			p [
				le [1, x] 2
			], p [
				eq [1, x] 1 ] ;

		"x=F(x)0", [{ Pol.var = x; Pol.lin = [Num.u, x]; Pol.cst = Num.u }],
			p [
				eq [1, x] 1
			], p [
				eq [-1, x] (-2) ];

		"x=F(x)1", [{ Pol.var = x; Pol.lin = [Num.u, x]; Pol.cst = Num.u }],
			p [
				le [1, x] 1
			], p [
				le [1, x] 2 ];

		"swap0", [{ Pol.var = x; Pol.lin = [Num.u, y]; Pol.cst = Num.z };
			{ Pol.var = y; Pol.lin = [Num.u, x]; Pol.cst = Num.z }],
			p [], p [];

		"swap1", [{ Pol.var = x; Pol.lin = [Num.u, y]; Pol.cst = Num.z };
			{ Pol.var = y; Pol.lin = [Num.u, x]; Pol.cst = Num.z }],
			p [
				eq [1, x] 1
			], p [
				eq [1, y] 1 ];

		"swap2", [{ Pol.var = x; Pol.lin = [Num.u, y]; Pol.cst = Num.z };
			{ Pol.var = y; Pol.lin = [Num.u, x]; Pol.cst = Num.z }],
			p [
				eq [1, y] 1
			], p [
				eq [1, x] 1 ];
	] in
	OUnit.TestList (List.map chk tcs)

(* Pol.widen *)
let widen_ts =
	let chk (t, p1, p2, pR) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		if Pol.isIncl p1 p2 = Pol.NoIncl then
			failwith "Pol_t.widen_ts"
		else
			let p = Pol.widen p1 p2 in
			let err =
				"p is different from expected pR" ^
				"\np:\n" ^ (Pol.prExt p) ^ "\npR:\n" ^ (Pol.prExt pR)
			in
			OUnit.assert_bool err (Pol.isEqSyn p pR)))
	in
	let tcs = [
		"halbwachs0", p [
			eq [1, y] 0;
			le [-1, x; 1, y] 0;
			le [1, x; -1, y] 1
		], p [
			le [-1, y] 0;
			le [-1, x; 1, y] 0;
			le [1, x; 1, y] 2
		], p [
			le [-1, y] 0;
			le [-1, x; 1, y] 0
		];

		"halbwachs1", p [
			eq [1, y] 0;
			le [-1, x] 0;
			le [1, x] 1
		], p [
			le [-1, y] 0;
			le [-1, x; 1, y] 0;
			le [1, x; 1, y] 2
		], p [
			le [-1, y] 0;
			le [-1, x; 1, y] 0
		]
	] in
	OUnit.TestList (List.map chk tcs)

let ts: OUnit.test = OUnit.TestList [
		OUnit.TestLabel ("add", add_ts);
		OUnit.TestLabel ("isIncl", isIncl_ts);
		OUnit.TestLabel ("project", project_ts);
		OUnit.TestLabel ("projectM", projectM_ts);
		OUnit.TestLabel ("setupJoin", setupJoin_ts);
		OUnit.TestLabel ("join", join_ts);
		OUnit.TestLabel ("assign", assign_ts);
		OUnit.TestLabel ("assignM", assignM_ts);
		OUnit.TestLabel ("widen", widen_ts)
	]
